import uuid
from copy import deepcopy
from functools import partial
from typing import List
from typing import Type
from typing import TypeVar

import pytest
import shapely.geometry
from shapely.geometry import Polygon

from paquo._utils import QuPathVersion
from paquo.classes import QuPathPathClass
from paquo.hierarchy import QuPathPathObjectHierarchy
from paquo.images import QuPathImageType
from paquo.pathobjects import QuPathPathAnnotationObject, _PathROIObject, QuPathPathDetectionObject
from paquo.projects import QuPathProject


@pytest.fixture(scope="function")
def empty_hierarchy():
    yield QuPathPathObjectHierarchy()


def test_initial_state(empty_hierarchy: QuPathPathObjectHierarchy):
    h = empty_hierarchy
    repr(h)
    assert h.is_empty
    assert h.root is not None  # root is auto populated
    assert len(h) == 0


_T = TypeVar('_T', bound=_PathROIObject)


def _make_polygons(obj_cls: Type[_T], amount: int) -> List[_T]:
    """returns a list of amount Path Objects"""
    path_objects = []
    for x in range(0, 10 * amount, 10):
        roi = shapely.geometry.Polygon.from_bounds(x, 0, x+5, 5)
        ao = obj_cls.from_shapely(roi)
        path_objects.append(ao)
    return path_objects


_make_polygon_annotations = partial(_make_polygons, QuPathPathAnnotationObject)
_make_polygon_detections = partial(_make_polygons, QuPathPathDetectionObject)


def test_attach_annotations(empty_hierarchy):
    h = empty_hierarchy
    # repr empty
    repr(h.annotations)

    annotations = _make_polygon_annotations(10)

    # add many
    h.annotations.update(annotations)

    # length
    assert len(h) == len(annotations)
    # contains
    assert annotations[3] in h.annotations
    assert 123 not in h.annotations
    # discard
    h.annotations.discard(annotations[7])
    assert len(h) == len(annotations) - 1
    # repr full
    repr(h.annotations)


def test_add_annotation_detection_tile(empty_hierarchy):
    empty_hierarchy.add_annotation(
        roi=shapely.geometry.Polygon.from_bounds(0, 0, 5, 5)
    )
    empty_hierarchy.add_detection(
        roi=shapely.geometry.Polygon.from_bounds(0, 0, 5, 5)
    )
    empty_hierarchy.add_tile(
        roi=shapely.geometry.Polygon.from_bounds(0, 0, 5, 5)
    )
    empty_hierarchy.add_cell(
        roi=shapely.geometry.Polygon.from_bounds(0, 0, 5, 5),
        nucleus_roi=shapely.geometry.Polygon.from_bounds(1.25, 1.25, 3.75, 3.75)
    )

def test_attach_detections(empty_hierarchy):
    h = empty_hierarchy
    detections = _make_polygon_detections(10)

    # add many
    h.detections.update(detections)

    # length
    assert len(h) == len(detections)
    # contains
    assert detections[3] in h.detections
    # discard
    h.detections.discard(detections[7])
    assert len(h) == len(detections) - 1


def test_annotations_detections_separation(empty_hierarchy):
    h = empty_hierarchy
    annotations = _make_polygon_annotations(5)
    detections = _make_polygon_detections(7)
    h.annotations.update(annotations)
    h.detections.update(detections)
    assert len(h.annotations) == 5
    assert len(h.detections) == 7


def test_add_cells(empty_hierarchy):
    h = empty_hierarchy
    roi = shapely.geometry.Polygon.from_bounds(0, 0, 5, 5)
    nucleus_roi = shapely.geometry.Polygon.from_bounds(1, 1, 4, 4)

    h.add_cell(roi=roi, nucleus_roi=nucleus_roi)

    assert len(h.detections) == 1
    assert len(h.cells) == 1

    cell = h.cells[0]
    assert shapely.equals(cell.roi, roi)
    assert shapely.equals(cell.nucleus_roi, nucleus_roi)


def test_geojson_roundtrip_via_geojson(empty_hierarchy):
    h = empty_hierarchy
    annotations = _make_polygon_annotations(10)

    h.annotations.update(annotations)
    assert len(h) == len(annotations)
    geojson = h.to_geojson()

    h.annotations.clear()
    assert len(h) == 0

    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        h.load_geojson("[]")

    h.load_geojson(geojson)
    assert len(h) == 10


TEST_ANNOTATION_POLYGON_VERSION_0_2_3 = [{
    'type': 'Feature',
    'id': 'PathAnnotationObject',
    'geometry': {
        'type': 'Polygon',
        'coordinates': [[
            [1000, 1300],
            [1011, 1420],
            [1120, 1430],
            [1060, 1380],
            [1000, 1300],
        ]]
    },
    'properties': {
        'classification': {
            'name': 'Tumor',
            'colorRGB': -3670016
        },
        'isLocked': False,
        'measurements': []
    }
}]

# qupath version v0.3.0-rc1 made a few changes to the geojson format
# 1. the 'id' key is removed and replaced by a property key 'object_type'
# 2. the property key 'measurements' is omitted if empty
TEST_ANNOTATION_POLYGON_VERSION_0_3_0_rc1_plus = [{
    'type': 'Feature',
    'geometry': {
        'type': 'Polygon',
        'coordinates': [[
            [1000, 1300],
            [1011, 1420],
            [1120, 1430],
            [1060, 1380],
            [1000, 1300],
        ]]
    },
    'properties': {
        'classification': {
            'name': 'Tumor',
            'colorRGB': -3670016
        },
        'isLocked': False,
        'object_type': 'annotation',
    }
}]


# qupath version v0.4.0 made changes to the geojson format
# 1. the 'id' key is added again but seems to be a UUID for each annotation
TEST_ANNOTATION_POLYGON_VERSION_0_4_0 = [{
    'type': 'Feature',
    'id': '7d254d72-a9a0-43a6-a455-3fa99e83b7af',
    'geometry': {
        'type': 'Polygon',
        'coordinates': [[
            [1000, 1300],
            [1011, 1420],
            [1120, 1430],
            [1060, 1380],
            [1000, 1300],
        ]]
    },
    'properties': {
        'classification': {
            'name': 'Tumor',
            'color': [200, 0, 0],
        },
        'isLocked': False,
        'objectType': 'annotation',
    }
}]


@pytest.fixture
def example_annotation(qupath_version):
    if qupath_version <= QuPathVersion("0.2.3"):
        yield deepcopy(TEST_ANNOTATION_POLYGON_VERSION_0_2_3)
    elif qupath_version < QuPathVersion("0.4.0"):
        yield deepcopy(TEST_ANNOTATION_POLYGON_VERSION_0_3_0_rc1_plus)
    else:
        yield deepcopy(TEST_ANNOTATION_POLYGON_VERSION_0_4_0)


def is_uuid(x):
    try:
        uuid.UUID(x)
    except ValueError:
        return False
    else:
        return True


@pytest.mark.parametrize(
    "input_annotation", [
        pytest.param(TEST_ANNOTATION_POLYGON_VERSION_0_2_3, id='v0.2.3'),
        pytest.param(TEST_ANNOTATION_POLYGON_VERSION_0_3_0_rc1_plus, id='v0.3.0rc1'),
        pytest.param(TEST_ANNOTATION_POLYGON_VERSION_0_4_0, id='v0.4.0'),
    ]
)
def test_geojson_roundtrip_via_annotations(empty_hierarchy, example_annotation, input_annotation, qupath_version):
    h = empty_hierarchy
    assert h.load_geojson(input_annotation)
    output = h.to_geojson()

    if qupath_version < QuPathVersion("0.4.0"):
        pass
    else:
        # these uuids are assigned randomly if they were missing in input
        out_id = output[0].pop("id", "")
        test_id = example_annotation[0].pop("id", "")
        assert is_uuid(out_id) and is_uuid(test_id)
        # isLocked is not exported
        del example_annotation[0]["properties"]["isLocked"]
    assert output == example_annotation


@pytest.fixture(scope='function')
def new_project(tmp_path):
    yield QuPathProject(tmp_path / "paquo-project", mode='x')


@pytest.fixture(scope='function')
def project_with_classes(new_project):
    new_project.path_classes = [
        QuPathPathClass("class0", color="#ff0000"),
        QuPathPathClass("class1", color="#00ff00"),
        QuPathPathClass("class2", color="#0000ff"),
    ]
    assert (
        new_project.path_classes
        and all(c.name.startswith('class') for c in new_project.path_classes)
    )
    yield new_project


NUM_ANNOTATIONS = 4

@pytest.fixture(scope='function')
def project_with_annotations(project_with_classes, svs_small):
    pth_cls, *_ = project_with_classes.path_classes

    num_annotations = NUM_ANNOTATIONS

    with project_with_classes as qp:
        entry = qp.add_image(
            svs_small, image_type=QuPathImageType.BRIGHTFIELD_H_E
        )

        for x in range(num_annotations):
            entry.hierarchy.add_annotation(
                roi=Polygon.from_bounds(0 + x, 0 + x, 1 + x, 1 + x),
                path_class=pth_cls,
            )
        assert len(entry.hierarchy.annotations) == num_annotations
    yield project_with_classes


def test_hierarchy_annotation_proxy_getitem(project_with_annotations):
    h = project_with_annotations.images[0].hierarchy

    assert len(h.annotations) == NUM_ANNOTATIONS
    assert len(h.annotations[NUM_ANNOTATIONS//2:]) == NUM_ANNOTATIONS//2
    assert len(h.annotations[[0, 3]]) == 2
    for idx in range(len(h.annotations)):
        _ = h.annotations[idx]


def test_add_to_existing_hierarchy(project_with_annotations):
    # create a project with an image and annotations
    from shapely.geometry import Point
    from paquo.projects import QuPathProject

    p = project_with_annotations.path
    num_annotations = len(project_with_annotations.images[0].hierarchy)
    del project_with_annotations

    # read project
    with QuPathProject(p, mode="a") as qp:
        entry1 = qp.images[0]

        assert len(entry1.hierarchy) == num_annotations
        entry1.hierarchy.add_annotation(
            roi=Point(2, 2)
        )
        assert len(entry1.hierarchy) == num_annotations + 1


def test_add_duplicate_to_hierarchy(project_with_annotations):
    """adding duplicate annotations works"""
    # create a project with an image and annotations
    from paquo.projects import QuPathProject

    p = project_with_annotations.path
    num_annotations = len(project_with_annotations.images[0].hierarchy)
    annotation = next(iter(project_with_annotations.images[0].hierarchy.annotations))
    del project_with_annotations

    # read project
    with QuPathProject(p, mode="a") as qp:
        entry1 = qp.images[0]
        entry1.hierarchy.add_annotation(
            roi=annotation.roi
        )
        assert len(entry1.hierarchy) == num_annotations + 1


def test_hierarchy_update_on_annotation_update(project_with_annotations):
    """updating annotation or detection objects triggers is_changed on hierarchy"""
    # create a project with an image and annotations
    from paquo.projects import QuPathProject
    from paquo.classes import QuPathPathClass

    p = project_with_annotations.path
    num_annotations = len(project_with_annotations.images[0].hierarchy)
    del project_with_annotations

    # test update
    with QuPathProject(p, mode="a") as qp:
        entry1 = qp.images[0]

        assert len(entry1.hierarchy) == num_annotations

        for annotation in entry1.hierarchy.annotations:
            annotation.update_path_class(QuPathPathClass("new"))
            assert entry1.is_changed()  # every update changes
            assert len(entry1.hierarchy) == num_annotations

        assert all(a.path_class.name == "new" for a in entry1.hierarchy.annotations)


def test_hierarchy_no_autoflush_annotation_update(project_with_annotations):
    """updating annotation or detection objects doesn't immediately trigger update"""
    from paquo.classes import QuPathPathClass

    with project_with_annotations as qp:
        entry1 = qp.images[0]
        with entry1.hierarchy.no_autoflush():
            for annotation in entry1.hierarchy.annotations:
                annotation.update_path_class(QuPathPathClass("new"))
                assert not entry1.is_changed()

        assert entry1.is_changed()
        assert all(a.path_class.name == "new" for a in entry1.hierarchy.annotations)


@pytest.fixture(scope='function')
def entry_with_annotations(project_with_annotations):
    # test update
    with project_with_annotations as qp:
        yield qp.images[0]


def test_add_incorrect_to_hierarchy(empty_hierarchy):
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        empty_hierarchy.annotations.add("abc")

    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        empty_hierarchy.annotations.discard("abc")


def test_serialize_to_ome(project_with_annotations):
    pytest.importorskip("ome_types")
    xml = project_with_annotations.images[0].hierarchy.to_ome_xml()

    assert xml.startswith("<OME")
    assert xml.strip().endswith("</OME>")


TEST_ANNOTATION_OBJECT_TYPE_TILE = {
    "features": [
        {
            "geometry": {
                "coordinates": [
                    [
                        [0, 23904],
                        [350, 23904],
                        [350, 24254],
                        [0, 24254],
                        [0, 23904],
                    ]
                ],
                "type": "Polygon"
            },
            "properties": {
                "isLocked": True,
                "measurements": {
                    "prob_notumor": 0.9998319149017334,
                    "prob_tumor": 0.00016804548795334995
                },
                "objectType": "tile"
            },
            "type": "Feature"
        }
    ],
    "type": "FeatureCollection"
}


def test_load_object_type_tile_geojson(empty_hierarchy):
    geojson = deepcopy(TEST_ANNOTATION_OBJECT_TYPE_TILE)["features"]
    empty_hierarchy.load_geojson(geojson)



import platform
import shutil
import sys
import tempfile
from contextlib import nullcontext
from operator import itemgetter
from pathlib import Path, PureWindowsPath, PurePosixPath
from typing import Dict

import pytest

from paquo.hierarchy import QuPathPathObjectHierarchy
from paquo.images import QuPathImageType, ImageProvider
from paquo.projects import QuPathProject


@pytest.fixture(scope='module')
def image_entry(svs_small):
    with tempfile.TemporaryDirectory(prefix='paquo-') as tmpdir:
        qp = QuPathProject(tmpdir, mode='x')
        entry = qp.add_image(svs_small)
        yield entry


@pytest.fixture(scope='function')
def removable_svs_small(svs_small):
    kw = {} if sys.version_info < (3, 10) else {"ignore_cleanup_errors": True}
    with tempfile.TemporaryDirectory(
        prefix='paquo-',
        **kw
    ) as tmpdir:
        new_path = Path(tmpdir) / svs_small.name
        shutil.copy(svs_small, new_path)
        yield new_path


@pytest.fixture(scope='function')
def project_with_removed_image(removable_svs_small):
    with tempfile.TemporaryDirectory(prefix='paquo-') as tmpdir:
        with QuPathProject(tmpdir, mode='x') as qp:
            _ = qp.add_image(removable_svs_small, image_type=QuPathImageType.BRIGHTFIELD_H_E)
            qp.save()
            path = qp.path
        try:
            removable_svs_small.unlink()
        except PermissionError:
            if platform.system() == "Windows":
                pytest.xfail("Windows QuPath==0.5.0 permission issue")
            else:
                raise
        yield path


@pytest.fixture(scope='function')
def project_with_removed_image_without_image_data(removable_svs_small):
    with tempfile.TemporaryDirectory(prefix='paquo-') as tmpdir:
        with QuPathProject(tmpdir, mode='x') as qp:
            _ = qp.add_image(removable_svs_small)
            qp.save()
            path = qp.path
        try:
            removable_svs_small.unlink()
        except PermissionError:
            if platform.system() == "Windows":
                pytest.xfail("Windows QuPath==0.5.0 permission issue")
            else:
                raise
        yield path


def test_image_entry_return_hierarchy(image_entry):
    assert isinstance(image_entry.hierarchy, QuPathPathObjectHierarchy)


def test_identifiers(image_entry):
    assert image_entry.entry_id == "1"  # first image...
    assert image_entry.image_name == "CMU-1-Small-Region.svs"
    # not changed yet.
    image_entry.image_name = "new_name"
    assert image_entry.image_name == "new_name"
    assert repr(image_entry)


def test_path(image_entry):
    assert image_entry.entry_path.is_dir()


def test_image_properties_from_image_server(image_entry):
    assert image_entry.width == 2220
    assert image_entry.height == 2967
    assert image_entry.num_channels == 3
    assert image_entry.num_timepoints == 1
    assert image_entry.num_z_slices == 1


def test_image_downsample_levels(image_entry):
    levels = [
        {'downsample': 1.0,
         'height': 2967,
         'width': 2220},
        # todo: when openslide can be used by qupath, this downsample level
        #   in the test image disappears. investigate...
        {'downsample': 3.865438534407666,
         'height': 768,
         'width': 574},
    ]
    assert (
        image_entry.downsample_levels == levels
        or image_entry.downsample_levels == levels[:1]
    )


def test_metadata_interface(image_entry):

    assert repr(image_entry.metadata)

    assert len(image_entry.metadata) == 0
    image_entry.metadata["test_key"] = "test_value"
    assert "test_key" in image_entry.metadata
    assert image_entry.metadata["test_key"] == "test_value"

    image_entry.metadata.update({
        "123": "123", "1": "abc"
    })
    assert len(image_entry.metadata) == 3
    assert "1" in image_entry.metadata
    assert "not-found" not in image_entry.metadata

    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        del image_entry.metadata[123]
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        _ = image_entry.metadata[123]
    with pytest.raises(KeyError):
        _ = image_entry.metadata['not-found']

    image_entry.metadata = {}
    assert dict(image_entry.metadata) == {}

    image_entry.metadata['a'] = '2'
    del image_entry.metadata['a']
    assert len(image_entry.metadata) == 0


# noinspection PyTypeChecker
def test_metadata_non_str_items(image_entry):

    with pytest.raises(TypeError):
        image_entry.metadata[1] = "abc"

    with pytest.raises(TypeError):
        image_entry.metadata["1"] = 123


def test_imagedata_saving_for_removed_images(project_with_removed_image):
    with QuPathProject(project_with_removed_image, mode='r+') as qp:
        entry = qp.images[0]
        assert (entry.entry_path / "data.qpdata").is_file()


def test_imagedata_saving_for_removed_images_without_type(project_with_removed_image_without_image_data):
    with QuPathProject(project_with_removed_image_without_image_data, mode='r+') as qp:
        entry = qp.images[0]
        # todo: check if we actually want this behavior.
        #   at least it's documented now
        assert not (entry.entry_path / "data.qpdata").is_file()


def test_readonly_recovery_hierarchy(project_with_removed_image_without_image_data):
    with QuPathProject(project_with_removed_image_without_image_data, mode='r+') as qp:
        entry = qp.images[0]
        assert repr(entry.hierarchy)


def test_readonly_recovery_image_server(project_with_removed_image):
    from paquo.java import compatibility
    if not compatibility.supports_image_server_recovery():
        pytest.skip(f"unsupported in {compatibility.version}")

    with QuPathProject(project_with_removed_image, mode='r') as qp:
        image_entry = qp.images[0]

        assert image_entry.height
        assert image_entry.width
        assert image_entry.num_channels
        assert image_entry.num_timepoints
        assert image_entry.num_z_slices
        assert image_entry.downsample_levels

        assert repr(image_entry.hierarchy)


def test_properties_interface(image_entry):
    from paquo.java import DefaultProject

    assert repr(image_entry.properties)

    assert len(image_entry.properties) == 1
    assert str(DefaultProject.IMAGE_ID) in image_entry.properties

    with pytest.raises(TypeError):
        image_entry.properties[1] = 123

    image_entry.properties["annotated"] = "no"

    assert image_entry.properties["annotated"] == "no"

    assert 123 not in image_entry.properties

    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        del image_entry.properties[123]
    with pytest.raises(KeyError):
        _ = image_entry.properties['not-found']
    with pytest.raises(TypeError):
        _ = image_entry.properties[123]

    image_entry.properties = {}
    assert dict(image_entry.properties) == {}


def test_description(image_entry):
    assert image_entry.description == ""
    image_entry.description = "abc"
    assert image_entry.description == "abc"


def test_image_type(image_entry):
    # initially unset
    assert image_entry.image_type == QuPathImageType.UNSET

    with pytest.raises(TypeError):
        image_entry.image_type = 123

    image_entry.image_type = QuPathImageType.BRIGHTFIELD_H_E
    assert image_entry.image_type == QuPathImageType.BRIGHTFIELD_H_E


TEST_URIS: Dict[str, Dict] = {
    "no-uri": {  # weird non-uri found in one project
        "uri": "\\\\SHARE\\site\\2020-01-01\\image 123-X,X.svs",
        "parts": None,
        "path": None,
        "exception": ValueError,
    },
    "network-share": {  # share on win
        "uri": "file:////SHARE/site/ABC/image.svs",
        "parts": ('ABC', 'image.svs'),
        "path": PureWindowsPath("//SHARE/site/", 'ABC', 'image.svs'),
        "exception": None,
    },
    "win-simple": {  # win-style uris
        "uri": "file:/C:/ABC/image.svs",
        "parts": ('ABC', 'image.svs'),
        "path": PureWindowsPath("C:/", 'ABC', 'image.svs'),
        "exception": None,
    },
    "win-spaces": {
        "uri": "file:/C:/ABC%20-%20ABC/image.svs",
        "parts": ('ABC - ABC', 'image.svs'),
        "path": PureWindowsPath("C:/", 'ABC - ABC', 'image.svs'),
        "exception": None,
    },
    "win-commas": {
        "uri": "file:/C:/ABC,ABC/image.svs",
        "parts": ('ABC,ABC', 'image.svs'),
        "path": PureWindowsPath("C:/", 'ABC,ABC', 'image.svs'),
        "exception": None,
    },
    "win-space-comma": {
        "uri": "file:/C:/ABC%20,%20ABC/image.svs",
        "parts": ('ABC , ABC', 'image.svs'),
        "path": PureWindowsPath("C:/", 'ABC , ABC', 'image.svs'),
        "exception": None,
    },
    "win-comb-01": {
        "uri": "file:/D:/2020-01-01/image.svs",
        "parts": ('2020-01-01', 'image.svs'),
        "path": PureWindowsPath("D:/", '2020-01-01', 'image.svs'),
        "exception": None,
    },
    "win-comb-02": {
        "uri": "file:/F:/2020-01-01/image-123-abc.svs",
        "parts": ('2020-01-01', 'image-123-abc.svs'),
        "path": PureWindowsPath("F:/", '2020-01-01', 'image-123-abc.svs'),
        "exception": None,
    },
    "win-comb-03": {
        "uri": "file:/U:/site/ABC%20ABC/image.svs",
        "parts": ('site', 'ABC ABC', 'image.svs'),
        "path": PureWindowsPath("U:/", 'site', 'ABC ABC', 'image.svs'),
        "exception": None,
    },
    "posix": {  # posix-style uri
        "uri": "file:/site/image.svs",
        "parts": ('site', 'image.svs'),
        "path": PurePosixPath("/", 'site', 'image.svs'),
        "exception": None,
    },
}


@pytest.mark.parametrize(
    "uri,exc", list(map(itemgetter("uri", "exception"), TEST_URIS.values())),
    ids=list(TEST_URIS.keys())
)
def test_image_provider_raise_if_invalid_uri(uri, exc):
    cm = pytest.raises(exc) if exc else nullcontext()
    with cm:
        ImageProvider.path_from_uri(uri)


@pytest.mark.parametrize(
    "uri,path", [(x["uri"], x["path"]) for x in TEST_URIS.values() if x["exception"] is None],
    ids=[k for k, x in TEST_URIS.items() if x["exception"] is None]
)
def test_image_provider_uri_from_path(uri, path):
    # noinspection PyTypeChecker
    new_uri = ImageProvider.uri_from_path(path)
    assert uri == new_uri
    # NOTE: the following test is of course weaker
    assert ImageProvider.compare_uris(uri, new_uri)


@pytest.mark.parametrize(
    "uri,path", [(x["uri"], x["path"]) for x in TEST_URIS.values() if x["exception"] is None],
    ids=[k for k, x in TEST_URIS.items() if x["exception"] is None]
)
def test_image_provider_path_from_uri(uri: str, path: Path):
    c_path = ImageProvider.path_from_uri(uri)
    assert c_path.is_absolute()
    assert type(c_path) is type(path)
    assert c_path.parts == path.parts


def test_image_provider_uri_from_relpath_and_abspath():
    with pytest.raises(ValueError):
        ImageProvider.uri_from_path(Path('./abc.svs'))

    p = '/abc.svs' if platform.system() != "Windows" else 'C:/abc.svs'
    assert ImageProvider.uri_from_path(Path(p)).startswith('file:')



from contextlib import nullcontext
from pathlib import Path

import pytest

# noinspection PyProtectedMember
from paquo.jpype_backend import start_jvm, find_qupath, qupath_jvm_info_from_qupath_dir, _conda_qupath_dir


def test_start_jvm_run_twice():
    assert start_jvm() == start_jvm()


def test_non_standard_qupath_installation(tmp_path):
    qupath_dir = Path(tmp_path)
    with pytest.raises(FileNotFoundError):
        qupath_jvm_info_from_qupath_dir(qupath_dir, [])


def test_conda_qupath_dir(monkeypatch, tmp_path):
    monkeypatch.delenv("CONDA_PREFIX", raising=False)
    assert _conda_qupath_dir() is None


def test_find_qupath(tmp_path):
    # prepare dirs
    opt = Path(tmp_path).absolute() / "opt"
    opt.mkdir()
    for qp in ["QuPath-123", "QuPath-234"]:
        (opt / qp).mkdir()

    with pytest.raises(FileNotFoundError):
        # qupath_dir provided by user must point to a valid qupath dir
        find_qupath(qupath_dir=str(tmp_path.absolute()))

    with pytest.raises(ValueError):
        # there is no spoon
        find_qupath(
            qupath_search_dirs=[str(opt), 'not-anything'],
            qupath_search_conda=False
        )

    with pytest.raises(ValueError):
        # search only one
        find_qupath(
            qupath_search_dirs=str(opt),
            qupath_search_conda=False
        )

    # first search others than try conda
    _d = _conda_qupath_dir()
    if _d is not None and _d.is_dir():
        cm = nullcontext()
    else:
        cm = pytest.raises(ValueError)  # pragma: no cover
    with cm:
        # search conda last
        find_qupath(qupath_dir=None, qupath_search_conda=True, qupath_prefer_conda=False)


def test_find_qupath_java_opts_as_string():
    with pytest.raises(ValueError):
        # search only one
        find_qupath(
            java_opts="-Djava.library.path='/123' -Djava.launcher.path='/'",
            qupath_search_conda=False
        )



"""Paquo OSX app shim example"""
from setuptools import setup

setup(
    app=["PaquoOpenQpZip.py"],
    data_files=[],
    options={
        "py2app": {
            "argv_emulation": 1,
        }
    },
    setup_requires=["py2app"],
)



from collections import ChainMap
from functools import partial
from io import StringIO
from xml.etree.ElementTree import Element
from xml.etree.ElementTree import ElementTree

__all__ = ["br", "div", "h3", "h4", "img", "p", "span", "css", "repr_html", "repr_svg", "rawhtml"]


def repr_html(obj, *args, **kwargs):
    try:
        # noinspection PyProtectedMember
        return obj._repr_html_(*args, **kwargs)
    except AttributeError:
        return repr(obj)


def repr_svg(obj, *args, **kwargs):
    try:
        # noinspection PyProtectedMember
        return obj._repr_svg_(*args, **kwargs)
    except AttributeError:
        return repr(obj)


def css(style_dict):
    return ";".join(f"{k}:{v}" for k, v in style_dict.items())


class _Tag(str):
    """convenience helper class for building html"""
    __slots__ = ('tag',)

    def __new__(cls, name, *tags, default_style=None, **attrs):
        # convert args
        text = attrs.pop("text", None)
        default_style = default_style or {}
        if 'style' in attrs or default_style:
            attrs['style'] = css(ChainMap(attrs.pop('style'), default_style))
        # create the xml tag
        tag = Element(name, attrib=attrs)
        if text is not None:
            tag.text = text
        for t in tags:
            tag.append(t.tag)
        # write the tag to str
        with StringIO() as out:
            ElementTree(tag).write(out, encoding='unicode', method='html')
            tag_str = out.getvalue()
        # initialize str and attach the xml tag
        str_obj = super().__new__(cls, tag_str)
        str_obj.tag = tag
        return str_obj


class _NonEscapingString(str):
    # prevent raw html being escaped by:
    # https://github.com/python/cpython/blob/0be7c216/Lib/xml/etree/ElementTree.py#L1033-L1047
    def __contains__(self, item):
        if item in "&<>":
            return False
        return super().__contains__(item)  # pragma: no cover


# noinspection PyPep8Naming
def rawhtml(text: str):
    """hacky way to prevent escaping the provided string"""
    raw_text = _NonEscapingString(text)
    return _Tag(None, text=raw_text)


br = partial(_Tag, "br")
div = partial(_Tag, "div")
h3 = partial(_Tag, "h3")
h4 = partial(_Tag, "h4", default_style={"margin-bottom": "0.5em"})
img = partial(_Tag, "img")
p = partial(_Tag, "p")
span = partial(_Tag, "span")



import string
from typing import NamedTuple
from typing import Tuple
from typing import Union

from paquo.java import ColorTools
from paquo.java import Integer

ColorTypeRGB = Tuple[int, int, int]
ColorTypeRGBA = Tuple[int, int, int, int]
ColorType = Union[ColorTypeRGB, ColorTypeRGBA, 'QuPathColor', str]


class QuPathColor(NamedTuple):
    """color representation in paquo

    >>> c = QuPathColor(128, 128, 160, alpha=240)
    """
    red: int
    green: int
    blue: int
    alpha: int = 255

    def is_valid(self) -> bool:
        """tests if a QuPathColor is valid

        (there are currently no validation checks performed on __init__)
        """
        for value in self.to_rgba():
            if not (0 <= value <= 255):
                # raise ValueError("val not an integer 0 <= val <= 255")
                return False
        return True

    def to_rgb(self) -> ColorTypeRGB:
        """convert to 3 * uint8 rgb tuple"""
        return self.red, self.green, self.blue

    def to_rgba(self) -> ColorTypeRGBA:
        """convert to 4 * uint8 rgba tuple"""
        return self.red, self.green, self.blue, self.alpha

    def to_mpl_rgba(self) -> Tuple[float, float, float, float]:
        """convert to 4 * float rgba tuple (mpl compatible)"""
        r, g, b, a = self.to_rgba()
        return r / 255.0, g / 255.0, b / 255.0, a / 255.0

    def to_hex(self) -> str:
        """convert to hex color. loses alpha."""
        r, g, b = self.to_rgb()
        return f"#{r:02x}{g:02x}{b:02x}"

    @classmethod
    def from_hex(cls, hex_color: str) -> 'QuPathColor':
        """convert from hex_color"""
        if (
            not isinstance(hex_color, str)
            or len(hex_color) != 7
            or hex_color[0] != "#"
            or any(c not in string.hexdigits for c in hex_color[1:])
        ):
            raise ValueError("requires a hexcolor #000000 - #ffffff")
        return cls(*(int(hex_color[i:i+2], 16) for i in (1, 3, 5)))

    def to_java_rgb(self) -> Integer:
        """"convert to the java rgb integer representation used by qupath"""
        return ColorTools.makeRGB(*self.to_rgb())

    @classmethod
    def from_java_rgb(cls, java_rgb: int) -> 'QuPathColor':
        """convert from java but ignore the alpha value in java_rgb"""
        if not isinstance(java_rgb, int):
            raise TypeError("requires an integer")
        r = int(ColorTools.red(java_rgb))
        g = int(ColorTools.green(java_rgb))
        b = int(ColorTools.blue(java_rgb))
        # noinspection PyArgumentList
        return cls(r, g, b)

    def to_java_rgba(self) -> Integer:
        """"convert to the java argb integer representation used by qupath"""
        return ColorTools.makeRGBA(*self.to_rgba())

    @classmethod
    def from_java_rgba(cls, java_rgba: int) -> 'QuPathColor':
        """converts a java integer color into a QuPathColor instance"""
        if not isinstance(java_rgba, int):
            raise TypeError("requires an integer")
        r = int(ColorTools.red(java_rgba))
        g = int(ColorTools.green(java_rgba))
        b = int(ColorTools.blue(java_rgba))
        a = int(ColorTools.alpha(java_rgba))
        # noinspection PyArgumentList
        return cls(r, g, b, a)

    def __repr__(self) -> str:
        if self.alpha != 255:
            return f"Color{self.to_rgba()}"
        return f"Color{self.to_rgb()}"

    def _repr_html_(self):
        from paquo._repr import div
        from paquo._repr import span
        a = self.alpha / 255.
        alpha = f"alpha={a:0.3f}" if self.alpha != 255 else ""
        return div(
            span(text=f"{self.__class__.__name__}: ", style={"font-weight": "bold"}),
            div(style={
                "display": "inline-block",
                "vertical-align": "text-top",
                "margin-right": "0.2em",
                "width": "1em",
                "height": "1em",
                "border": "1px solid black",
                "opacity": str(a),
                "background": self.to_hex()
            }),
            span(text=f"{self.to_hex()} {alpha}"),
        )

    @classmethod
    def from_any(cls, value: ColorType) -> 'QuPathColor':
        """try creating a QuPathColor from all supported types"""
        if isinstance(value, QuPathColor):
            return cls(*value.to_rgba())
        elif isinstance(value, (tuple, list)):
            return cls(*value)
        elif isinstance(value, str):
            return cls.from_hex(value)
        else:
            raise TypeError("can't convert to QuPathColor")



import tempfile
import unittest.mock
from pathlib import Path

import pytest
import shapely.geometry

from paquo._repr import repr_html, repr_svg
from paquo.colors import QuPathColor
from paquo.images import QuPathProjectImageEntry
from paquo.pathobjects import QuPathPathAnnotationObject
from paquo.projects import QuPathProject


@pytest.fixture(scope='function')
def new_project(tmp_path):
    yield QuPathProject(tmp_path / "paquo-project", mode='x')


@pytest.fixture(scope='module')
def image_entry(svs_small):
    with tempfile.TemporaryDirectory(prefix='paquo-') as tmpdir:
        qp = QuPathProject(tmpdir, mode='x')
        entry = qp.add_image(svs_small)
        yield entry


def test_repr_helper_fallback():
    obj_without_ipynb_repr = "a"
    assert repr(obj_without_ipynb_repr) == repr_html(obj_without_ipynb_repr)
    assert repr(obj_without_ipynb_repr) == repr_svg(obj_without_ipynb_repr)


def test_ipython_repr(new_project):
    assert new_project._repr_html_()


def test_pathobject_repr():
    p = QuPathPathAnnotationObject.from_shapely(
        roi=shapely.geometry.Point(1, 2)
    )
    assert repr_html(p)


def test_color_repr():
    c = QuPathColor.from_hex("#123456")
    assert repr_html(c)


def test_hierarchy_repr(image_entry):
    assert repr_html(image_entry.hierarchy)


def test_image_repr(image_entry, monkeypatch):
    assert repr_html(image_entry)
    assert repr_html(image_entry, compact=True)



import json
import math
from collections.abc import MutableMapping
from functools import partial
from typing import Callable
from typing import Iterator
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union

from shapely.geometry import shape
from shapely.geometry.base import BaseGeometry
from shapely.wkb import dumps as shapely_wkb_dumps
from shapely.wkb import loads as shapely_wkb_loads

from paquo._utils import cached_property
from paquo.classes import QuPathPathClass
from paquo.java import ROI
from paquo.java import GeometryTools
from paquo.java import GsonTools
from paquo.java import PathAnnotationObject
from paquo.java import PathCellObject
from paquo.java import PathDetectionObject
from paquo.java import PathObjects
from paquo.java import PathROIObject
from paquo.java import PathTileObject
from paquo.java import String
from paquo.java import WKBReader
from paquo.java import WKBWriter

__all__ = [
    "fix_geojson_geometry",
    "BaseGeometry",
    "PathROIObjectType",
    "QuPathPathAnnotationObject",
    "QuPathPathDetectionObject",
    "QuPathPathTileObject",
    "QuPathPathCellObject",
]


def _shapely_geometry_to_qupath_roi(geometry: BaseGeometry, image_plane=None) -> ROI:
    """convert a shapely geometry into a qupath ROI

    uses well known binary WKB as intermediate representation

    todo: expose image plane settings and provide pythonic interface to it
    """
    wkb_bytes = shapely_wkb_dumps(geometry)
    jts_geometry = WKBReader().read(wkb_bytes)
    return GeometryTools.geometryToROI(jts_geometry, image_plane)


def _qupath_roi_to_shapely_geometry(roi) -> BaseGeometry:
    """convert a qupath ROI to a shapely geometry

    uses well known binary WKB as intermediate representation

    note: this loses the image plane information
    """
    jts_geometry = GeometryTools.roiToGeometry(roi)
    wkb_bytearray = WKBWriter(2).write(jts_geometry)
    return shapely_wkb_loads(bytes(wkb_bytearray))


def fix_geojson_geometry(geometry: dict) -> dict:
    """try to fix a provided geojson geometry via buffering"""
    s = shape(geometry)
    if not s.is_valid:
        # attempt to fix
        s = s.buffer(0, 1)
        if not s.is_valid:
            s = s.buffer(0, 1)
            if not s.is_valid:
                raise ValueError("invalid geometry")
    return s.__geo_interface__  # type: ignore


class _MeasurementList(MutableMapping):

    def __init__(
        self,
        measurement_list,
        *,
        update_callback: Optional[Callable[[], None]] = None
    ):
        self._measurement_list = measurement_list
        self._update_callback = update_callback

    def __setitem__(self, k: str, v: float) -> None:
        if not isinstance(v, float):
            raise TypeError(f"value must be float, got: {type(v).__name__}")
        self._measurement_list.putMeasurement(k, v)
        if self._update_callback:
            self._update_callback()

    def __delitem__(self, v: str) -> None:
        if v not in self:
            raise KeyError(v)
        self._measurement_list.removeMeasurements(v)
        if self._update_callback:
            self._update_callback()

    def __getitem__(self, k: Union[str, int]) -> float:
        if not isinstance(k, (int, str)):
            raise KeyError(f"unsupported key of type {type(k)}")
        value = float(self._measurement_list.getMeasurementValue(k))
        if math.isnan(value) and k not in self:
            raise KeyError(k)
        return value

    def __contains__(self, item: object) -> bool:
        if not isinstance(item, str):
            return False
        return bool(self._measurement_list.containsNamedMeasurement(item))

    def __len__(self) -> int:
        return int(self._measurement_list.size())

    def __iter__(self) -> Iterator[str]:
        return iter(map(str, self._measurement_list.getMeasurementNames()))

    def clear(self) -> None:
        self._measurement_list.clear()
        if self._update_callback:
            self._update_callback()

    def __repr__(self):
        return f"Measurements({repr(dict(self))})"

    def __str__(self):
        return str(dict(self))

    def to_records(self):
        return [{'name': name, 'value': value} for name, value in self.items()]


# noinspection PyTypeChecker
PathROIObjectType = TypeVar('PathROIObjectType', bound='_PathROIObject')


class _PathROIObject:
    """internal base class for PathObjects"""

    # must be provided in subclass
    java_class: Type[PathROIObject]
    java_class_factory: Callable[..., PathROIObject]

    def __init__(
        self,
        java_object: PathROIObject,
        *,
        update_callback: Optional[Callable[[PathROIObjectType], None]] = None
    ) -> None:
        """instantiate using classmethods: `from_shapely`, `from_geojson`"""
        self.java_object = java_object
        self._update_callback = update_callback

    @classmethod
    def from_shapely(cls: Type[PathROIObjectType],
                     roi: BaseGeometry,
                     path_class: Optional[QuPathPathClass] = None,
                     measurements: Optional[dict] = None,
                     *,
                     path_class_probability: float = math.nan) -> "PathROIObjectType":
        """create a Path Object from a shapely shape

        Parameters
        ----------
        roi:
            a shapely shape as the region of interest of the annotation
        path_class:
            a paquo QuPathPathClass to mark the annotation type
        measurements:
            dict holding static measurements for annotation object
        path_class_probability:
            keyword only argument defining the probability of the class
            (default NaN)

        """
        if not isinstance(roi, BaseGeometry):
            raise TypeError("roi needs to be an instance of shapely.geometry.base.BaseGeometry")
        qupath_roi = _shapely_geometry_to_qupath_roi(roi)
        qupath_path_class = path_class.java_object if path_class is not None else None
        # fixme: should create measurements here and pass instead of None
        java_obj = cls.java_class_factory(qupath_roi, qupath_path_class, None)
        if not math.isnan(path_class_probability):
            java_obj.setPathClass(java_obj.getPathClass(), path_class_probability)
        obj = cls(java_obj)
        if measurements is not None:
            obj.measurements.update(measurements)
        return obj

    @classmethod
    def from_geojson(cls: Type[PathROIObjectType], geojson) -> PathROIObjectType:
        """create a new Path Object from geojson"""
        gson = GsonTools.getInstance()
        java_obj = gson.fromJson(String(json.dumps(geojson)), cls.java_class)
        return cls(java_obj)

    def to_geojson(self) -> dict:
        """convert the annotation object to geojson"""
        gson = GsonTools.getInstance()
        geojson = gson.toJson(self.java_object)
        return dict(json.loads(str(geojson)))

    @property
    def path_class(self) -> Optional[QuPathPathClass]:
        """the annotation path class"""
        pc = self.java_object.getPathClass()
        if not pc:
            return None
        return QuPathPathClass.from_java(pc)

    @property
    def path_class_probability(self) -> float:
        """the annotation path class probability"""
        return float(self.java_object.getClassProbability())

    def update_path_class(self: PathROIObjectType, pc: Optional[QuPathPathClass], probability: float = math.nan) -> None:
        """updating the class or probability has to be done via this method"""
        if not (pc is None or isinstance(pc, QuPathPathClass)):
            raise TypeError("requires QuPathPathClass")
        else:
            pc = pc if pc is None else pc.java_object
        self.java_object.setPathClass(pc, probability)
        if self._update_callback:
            self._update_callback(self)  # type: ignore[arg-type]

    @property
    def locked(self) -> bool:
        """lock state of the annotation"""
        return bool(self.java_object.isLocked())

    @locked.setter
    def locked(self: PathROIObjectType, value: bool) -> None:
        self.java_object.setLocked(value)
        if self._update_callback:
            self._update_callback(self)  # type: ignore[arg-type]

    @property
    def is_editable(self) -> bool:
        """can the annotation be edited in the qupath UI"""
        return bool(self.java_object.isEditable())

    @property
    def level(self) -> int:
        """the annotation's level"""
        return int(self.java_object.getLevel())

    @property
    def name(self) -> Optional[str]:
        """an optional name for the annotation"""
        name = self.java_object.getName()
        if name is None:
            return None
        return str(name)

    @name.setter
    def name(self: PathROIObjectType, name: Union[str, None]) -> None:
        if name is not None:
            name = String(name)
        self.java_object.setName(name)
        if self._update_callback:
            self._update_callback(self)  # type: ignore[arg-type]

    @property
    def parent(self: PathROIObjectType) -> Optional[PathROIObjectType]:
        """the annotation object's parent annotation object"""
        parent = self.java_object.getParent()
        if not parent:
            return None
        # fixme: Is this true? Or do we need to dynamically cast to the right subclass
        return self.__class__(parent)

    @property
    def roi(self) -> BaseGeometry:
        """the roi as a shapely shape"""
        roi = self.java_object.getROI()
        return _qupath_roi_to_shapely_geometry(roi)

    def update_roi(self: PathROIObjectType, geometry: BaseGeometry) -> None:
        """update the roi of the annotation"""
        roi = _shapely_geometry_to_qupath_roi(geometry)
        self.java_object.setROI(roi)
        if self._update_callback:
            self._update_callback(self)  # type: ignore[arg-type]

    @cached_property
    def measurements(self):
        if self._update_callback:
            cb = partial(self._update_callback, self)
        else:
            cb = None
        return _MeasurementList(
            self.java_object.getMeasurementList(),
            update_callback=cb,
        )

    def __repr__(self):
        name = self.name
        path_class = self.path_class
        roi = self.roi
        out = []
        if name:
            out.append(f'name="{name}"')
        if path_class:
            out.append(f'class="{path_class.name}"')
        if roi:
            out.append(f'roi={roi.geom_type}')
        return f"{type(self).__name__}({' '.join(out)})"

    def _repr_html_(self):
        from paquo._repr import br
        from paquo._repr import div
        from paquo._repr import h4
        from paquo._repr import p
        from paquo._repr import rawhtml
        from paquo._repr import repr_svg
        from paquo._repr import span

        obj_class_name = self.__class__.__name__
        if obj_class_name.startswith('QuPath'):
            obj_class_name = obj_class_name[6:]

        name = self.name or "N/A"
        path_class = self.path_class
        path_class_name = path_class.name if path_class else "N/A"
        roi = self.roi
        if hasattr(roi, '_repr_svg_'):
            roi_tag = span(rawhtml(repr_svg(roi)), style={"vertical-align": "text-top"})
        else:
            roi_tag = span(text=roi.wkt)  # pragma: no cover

        return div(
            h4(text=f"{obj_class_name}:", style={"margin-top": "0"}),
            p(
                span(text="name: ", style={"font-weight": "bold"}),
                span(text=name),
                br(),
                span(text="path_class: ", style={"font-weight": "bold"}),
                span(text=path_class_name),
                br(),
                span(text="roi_type: ", style={"font-weight": "bold"}),
                span(text=roi.geom_type),
                br(),
                span(text="roi: ", style={"font-weight": "bold"}),
                roi_tag,
                style={"margin": "0.5em"},
            ),
        )


class QuPathPathAnnotationObject(_PathROIObject):

    java_class = PathAnnotationObject
    java_class_factory = PathObjects.createAnnotationObject

    @property
    def description(self) -> Optional[str]:
        """an optional description for the annotation"""
        desc = self.java_object.getDescription()
        return str(desc) if desc is not None else None

    @description.setter
    def description(self, value: str):
        if not isinstance(value, str):
            raise TypeError("requires a str")
        self.java_object.setDescription(String(value))


class QuPathPathDetectionObject(_PathROIObject):

    java_class = PathDetectionObject
    java_class_factory = PathObjects.createDetectionObject


class QuPathPathTileObject(QuPathPathDetectionObject):

    java_class = PathTileObject
    java_class_factory = PathObjects.createTileObject


class QuPathPathCellObject(QuPathPathDetectionObject):

    java_class = PathCellObject
    java_class_factory = PathObjects.createCellObject

    @property
    def nucleus_roi(self) -> BaseGeometry:
        """the nucleus roi as a shapely shape"""
        roi = self.java_object.getNucleusROI()
        return _qupath_roi_to_shapely_geometry(roi)

    @classmethod
    def from_shapely(cls,
                     roi: BaseGeometry,
                     path_class: Optional[QuPathPathClass] = None,
                     measurements: Optional[dict] = None,
                     *,
                     path_class_probability: float = math.nan,
                     nucleus_roi: Optional[BaseGeometry] = None) -> "QuPathPathCellObject":
        """create a Path Object from a shapely shape

        Parameters
        ----------
        roi:
            a shapely shape as the region of interest of the annotation
        path_class:
            a paquo QuPathPathClass to mark the annotation type
        measurements:
            dict holding static measurements for annotation object
        path_class_probability:
            keyword only argument defining the probability of the class
            (default NaN)
        nucleus_roi:
            a roi for a nucleus

        """
        if not isinstance(roi, BaseGeometry):
            raise TypeError("roi needs to be an instance of shapely.geometry.base.BaseGeometry")
        if not isinstance(nucleus_roi, BaseGeometry):
            raise TypeError("nucleus_roi needs to be an instance of shapely.geometry.base.BaseGeometry")

        qupath_path_class = path_class.java_object if path_class is not None else None
        qupath_roi = _shapely_geometry_to_qupath_roi(roi)
        qupath_additional_roi = _shapely_geometry_to_qupath_roi(nucleus_roi)
        java_obj = cls.java_class_factory(qupath_roi, qupath_additional_roi, qupath_path_class, None)

        if not math.isnan(path_class_probability):
            java_obj.setPathClass(java_obj.getPathClass(), path_class_probability)
        obj = cls(java_obj)
        if measurements is not None:
            obj.measurements.update(measurements)
        return obj



"""example showing how to read annotations from an existing project"""
from pathlib import Path
from paquo.projects import QuPathProject

EXAMPLE_PROJECT = Path(__file__).parent.absolute() / "projects" / "example_01_project"

# read the project and raise Exception if it's not there
with QuPathProject(EXAMPLE_PROJECT, mode='r') as qp:
    print("opened", qp.name)
    # iterate over the images
    for image in qp.images:
        # annotations are accessible via the hierarchy
        annotations = image.hierarchy.annotations

        print("Image", image.image_name, "has", len(annotations), "annotations.")
        for annotation in annotations:
            # annotations are paquo.pathobjects.QuPathPathAnnotationObject instances
            # their ROIs are accessible as shapely geometries via the .roi property
            print("> class:", annotation.path_class.name, "roi:", annotation.roi)

    print("done")



import sys
import tempfile
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List

from dynaconf import Dynaconf
from dynaconf import Validator
from dynaconf import loaders
from dynaconf.base import Settings
from dynaconf.utils import files as _files
from dynaconf.utils.boxing import DynaBox

if sys.version_info >= (3, 9):
    from importlib.resources import as_file
    from importlib.resources import files

    def importlib_resources_path(package, resource):
        return as_file(files(package).joinpath(resource))

else:
    from importlib.resources import path as importlib_resources_path


PAQUO_CONFIG_FILENAME = '.paquo.toml'

_PAQUO_FIX_CONFIG_KEYS = {
    # recover prefixed keys due to dynaconf weirdness
    "_qupath_dir",
    "_qupath_search_dirs",
    "_qupath_search_dir_regex",
    "_qupath_search_conda",
    "_qupath_prefer_conda",
    "_java_opts",
    "_safe_truncate",
    "_mock_backend",
    "_cli_force_log_level_error",
    "_warn_microsoft_store_python",
}


def to_kwargs(s: "Settings | DynaBox") -> Dict[str, Any]:
    """convert dynaconf settings to lowercase"""
    dct = s.to_dict()
    out = {}
    for key, value in dct.items():
        k = key.lower()
        if k in _PAQUO_FIX_CONFIG_KEYS:
            k = k[1:]
        out[k] = value
    return out


def to_toml(s: Settings) -> str:
    """convert dynaconf settings to a toml str"""
    # we'll use the current public dynaconf api so that we don't need to import our own
    # toml... that means we need to go via a temporary file to dump the current config,
    # because dynaconf doesn't support it otherwise...
    # github.com/rochacbruno/dynaconf/blob/68df27d2/dynaconf/loaders/toml_loader.py#L56
    data = DynaBox(s.as_dict(internal=False))

    # we create a temporary dir and write to a toml file
    # note: this is to workaround the fact that using NamedTemporaryFile will
    #   error under windows due to loaders.write calling open on the file
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = str(Path(tmpdir) / ".paquo.temporary.toml")  # suffix determines loader
        loaders.write(fn, to_kwargs(data))
        with open(fn) as f:
            output = f.read()
    return output


def _get_settings() -> Dynaconf:
    """load default settings instance"""
    with importlib_resources_path("paquo", ".paquo.defaults.toml") as default_config:

        settings = Dynaconf(
            envvar_prefix="PAQUO",
            settings_file=[PAQUO_CONFIG_FILENAME],
            root_path=Path.cwd(),
            core_loaders=['TOML'],
            preload=[str(default_config.absolute())],
            validators=[
                Validator("java_opts", is_type_of=(list, tuple, str)),
                Validator("qupath_search_dirs", is_type_of=(list, tuple, str)),
                Validator("qupath_search_conda", is_type_of=(bool, int), is_in=(0, 1)),
                Validator("qupath_prefer_conda", is_type_of=(bool, int), is_in=(0, 1)),
                Validator("safe_truncate", is_type_of=(bool, int), is_in=(0, 1)),
                Validator("mock_backend", is_type_of=(bool, int), is_in=(0, 1)),
                Validator("cli_force_log_level_error", is_type_of=(bool, int), is_in=(0, 1)),
            ]
        )
    return settings


settings = _get_settings()


def get_searchtree() -> List[str]:
    """return the current search tree for the settings"""
    if not settings.configured:
        settings.configure()  # pragma: no cover
    # note: SEARCHTREE is updated after configure
    searchtree: List[str] = getattr(_files, 'SEARCHTREE', [])
    return searchtree



import importlib.util
import shutil
import tempfile
from contextlib import contextmanager
from pathlib import Path

import pytest
import shapely.geometry

from paquo.images import ImageProvider
from paquo.projects import QuPathProject
from paquo._utils import cached_property


@pytest.fixture(scope='module')
def project_and_changes(svs_small):
    with tempfile.TemporaryDirectory(prefix='paquo-') as tmpdir:
        qp = QuPathProject(tmpdir, mode='x')
        entry = qp.add_image(svs_small)
        entry.hierarchy.add_annotation(
            roi=shapely.geometry.Point(1, 2)
        )
        qp.save()
        project_path = qp.path.parent
        del qp

        last_changes = {}
        for file in project_path.glob("**/*.*"):
            p = str(file.absolute())
            last_changes[p] = file.stat().st_mtime

        yield project_path, last_changes


@pytest.fixture(scope='function')
def copy_svs_small(svs_small):
    with tempfile.TemporaryDirectory(prefix='paquo-') as tmpdir:
        new_path = Path(tmpdir) / svs_small.name
        shutil.copy(svs_small, new_path)
        yield new_path


@pytest.fixture(scope="function")
def readonly_project(project_and_changes):
    project_path, changes = project_and_changes
    qp = QuPathProject(project_path, mode="r")
    qp.__changes = changes
    yield qp


def iter_readonly_properties(obj):
    cls = obj.__class__
    for prop in dir(cls):
        cls_prop = getattr(cls, prop)
        if isinstance(cls_prop, property) and cls_prop.fset is None:
            yield prop
        if isinstance(cls_prop, cached_property):
            yield prop


@contextmanager
def assert_no_modification(qp):
    ctime, mtime = qp.timestamp_creation, qp.timestamp_modification
    yield qp
    project_path = qp.path.parent
    files = project_path.glob("**/*.*")
    assert files
    for file in files:
        p = str(file.absolute())
        assert qp.__changes.get(p, None) == file.stat().st_mtime, f"{str(file.relative_to(project_path))} was modified"
    assert qp.timestamp_creation == ctime
    assert qp.timestamp_modification == mtime


def test_fixture(readonly_project):
    pass


class _Accessor:
    def __init__(self, instance):
        self._if = set(filter(lambda x: not x.startswith('_'), dir(type(instance))))
        self._i = instance

    def setattr(self, item, value):
        self._if.discard(item)
        setattr(self._i, item, value)

    def callmethod(self, method, *args, ignore_exception=False, **kwargs):
        self._if.discard(method)
        try:
            return getattr(self._i, method)(*args, **kwargs)
        except BaseException as e:
            if ignore_exception:
                pass
            else:
                raise e

    def unused_public_interface(self):
        return self._if


def test_project_attrs_and_methods(readonly_project, copy_svs_small):
    with assert_no_modification(readonly_project) as qp:
        a = _Accessor(qp)

        assert qp._readonly

        # these are readonly anyways
        for ro_prop in iter_readonly_properties(qp):
            with pytest.raises(AttributeError):
                a.setattr(ro_prop, "abc")
        #with pytest.raises(AttributeError):
        #    a.setattr("images", [])

        # these dont do anything
        a.callmethod("is_readable")

        # modifiable: These should all raise AttributeError in readonly mode
        with pytest.raises(AttributeError):
            a.setattr("path_classes", ())

        # These should raise an IOError when readonly
        with pytest.raises(IOError):
            a.callmethod("add_image", copy_svs_small, allow_duplicates=True)
        with pytest.raises(IOError):
            a.callmethod("save")
        with pytest.raises(IOError):
            image_entry = qp.images[0]
            a.callmethod("remove_image", image_entry)

        # test that we can reassign uri's even in readonly mode
        cur_uri = qp.images[0].uri
        new_uri = ImageProvider.uri_from_path(copy_svs_small)
        assert cur_uri != new_uri
        # test that this does not change the project
        a.callmethod("update_image_paths", uri2uri={cur_uri: new_uri})
        assert qp.images[0].uri == new_uri

        # make sure everything is covered in case we extend the classes later
        assert not a.unused_public_interface()


def test_images_attrs_methods_readonly(readonly_project):
    with assert_no_modification(readonly_project) as qp:
        image = qp.images[0]
        i = _Accessor(image)

        assert image._readonly

        # readonly properties
        for ro_prop in iter_readonly_properties(image):
            with pytest.raises(AttributeError):
                i.setattr(ro_prop, "abc")

        # test writable properties
        with pytest.raises(AttributeError):
            i.setattr("description", "abc")
        with pytest.raises(AttributeError):
            i.setattr("image_name", "abc")
        with pytest.raises(AttributeError):
            i.setattr("image_type", "abc")

        # these do nothing
        i.callmethod("is_changed")
        i.callmethod("is_readable")

        # these need to be blocked
        with pytest.raises(AttributeError):
            i.setattr("metadata", {})
        with pytest.raises(AttributeError):
            i.setattr("properties", {})

        # methods that are not allowed
        with pytest.raises(IOError):
            i.callmethod("save")

        assert not i.unused_public_interface()


def test_images_metadata_and_properties(readonly_project):

    with assert_no_modification(readonly_project) as qp:
        image = qp.images[0]
        assert image._readonly

        with pytest.raises(AttributeError):
            image.metadata[12] = 3
        with pytest.raises(AttributeError):
            image.metadata.clear()
        with pytest.raises(AttributeError):
            del image.metadata[12]

        with pytest.raises(AttributeError):
            image.properties[12] = 3
        with pytest.raises(AttributeError):
            image.properties.clear()
        with pytest.raises(AttributeError):
            del image.properties[12]


def test_hierarchy(readonly_project):

    OME_NOT_INSTALLED = importlib.util.find_spec("ome_types") is None

    with assert_no_modification(readonly_project) as qp:
        image = qp.images[0]
        hierarchy = image.hierarchy
        assert hierarchy._readonly

        h = _Accessor(hierarchy)

        # readonly properties
        for ro_prop in iter_readonly_properties(hierarchy):
            with pytest.raises(AttributeError):
                h.setattr(ro_prop, "abc")

        # these do nothing
        h.callmethod("to_geojson")
        h.callmethod("to_ome_xml", ignore_exception=OME_NOT_INSTALLED)

        # these are not allowed in readonly
        with pytest.raises(IOError):
            h.callmethod("add_annotation", '--placeholder--')
        with pytest.raises(IOError):
            h.callmethod("add_detection", '--placeholder--')
        with pytest.raises(IOError):
            h.callmethod("add_tile", '--placeholder--')
        with pytest.raises(IOError):
            h.callmethod("add_cell", '--placeholder--', nucleus_roi='--placeholder--')
        with pytest.raises(IOError):
            h.callmethod("load_geojson", '--placeholder--')

        # autoflush has no influence
        h.setattr("autoflush", False)
        h.callmethod("no_autoflush")
        h.callmethod("flush")

        assert not h.unused_public_interface()


def test_hierarchy_annotations_detections(readonly_project):
    with assert_no_modification(readonly_project) as qp:
        image = qp.images[0]
        hierarchy = image.hierarchy
        assert hierarchy._readonly

        with pytest.raises(IOError):
            hierarchy.annotations.add(1)
        with pytest.raises(IOError):
            hierarchy.annotations.discard(1)
        with pytest.raises(IOError):
            hierarchy.detections.add(1)
        with pytest.raises(IOError):
            hierarchy.detections.discard(1)



"""prepare_resources.py for paquo/examples

note: this code initializes data and projects needed for
  the examples in the paquo repository. Some of it is
  copy&pasta from the pytest fixtures.

"""
import io
import itertools
import json
import pathlib
import shutil
import urllib.request

EXAMPLES_DIR = pathlib.Path(__file__).parent.absolute()
IMAGES_DIR = EXAMPLES_DIR / "images"
PROJECTS_DIR = EXAMPLES_DIR / "projects"
DATA_DIR = EXAMPLES_DIR / "data"

for example_dir in [IMAGES_DIR, PROJECTS_DIR, DATA_DIR]:
    example_dir.mkdir(parents=True, exist_ok=True)


def download_image(copies=1):
    """download the smallest aperio test image svs"""
    # openslide aperio test image
    images_base_url = "http://openslide.cs.cmu.edu/download/openslide-testdata/Aperio/"
    small_image = "CMU-1-Small-Region.svs"
    # download svs from openslide test images
    url = images_base_url + small_image
    for idx in range(copies):
        img_fn = IMAGES_DIR / f"image_{idx}.svs"
        yield img_fn


def prepare_example_resources():
    """build an example project"""
    from paquo.projects import QuPathProject
    from paquo.images import QuPathImageType
    from paquo.classes import QuPathPathClass
    from shapely.geometry import Polygon

    # download all images
    print(">>> downloading images...")
    images = list(download_image(copies=3))
    # remove the example_project
    example_project_dir = PROJECTS_DIR / "example_01_project"
    shutil.rmtree(example_project_dir, ignore_errors=True)

    print(">>> creating project...")
    with QuPathProject(example_project_dir, mode="x") as qp:
        for img_fn in images:
            qp.add_image(img_fn, image_type=QuPathImageType.BRIGHTFIELD_H_E)
        qp.path_classes = map(QuPathPathClass, ["myclass_0", "myclass_1", "myclass_2"])
        for idx, image in enumerate(qp.images):
            image.metadata['image_number'] = str(1000 + idx)

        def _get_bounds(idx_x, idx_y):
            cx = 200 * idx_x + 100
            cy = 200 * idx_y + 100
            return cx - 50, cy - 50, cx + 50, cy + 50

        image_0 = qp.images[0]
        for x, y in itertools.product(range(4), repeat=2):
            roi = Polygon.from_bounds(*_get_bounds(x, y))
            image_0.hierarchy.add_annotation(roi, path_class=QuPathPathClass("myclass_1"))

        with open(DATA_DIR / "annotations.geojson", "w") as f:
            json.dump(image_0.hierarchy.to_geojson(), f)

    print(">>> done.")


if __name__ == "__main__":
    prepare_example_resources()



"""example showing how to create a project with image metadata"""
from pathlib import Path
from paquo.projects import QuPathProject
from paquo.images import QuPathImageType

EXAMPLE_PROJECT = Path(__file__).parent.absolute() / "projects" / "example_04_project"
EXAMPLE_IMAGE_DIR = Path(__file__).parent.absolute() / "images"

METADATA = {
    "image_0.svs": {
        "annotator": "Alice",
        "status": "finished",
        "diagnosis": "healthy",
    },
    "image_1.svs": {
        "annotator": "Bob",
        "status": "started",
        "diagnosis": "sick",
    },
    "image_2.svs": {
        "annotator": "Chuck",
        "status": "waiting",
        "diagnosis": "unknown",
    },
}

# create a the new project
with QuPathProject(EXAMPLE_PROJECT, mode='x') as qp:
    print("created", qp.name)

    for image_fn, metadata in METADATA.items():
        entry = qp.add_image(
            EXAMPLE_IMAGE_DIR / image_fn,
            image_type=QuPathImageType.BRIGHTFIELD_H_E
        )
        # entry.metadata is a dict-like proxy:
        # > entry.metadata[key] = value
        # > entry.metadata.update(new_dict)
        # > etc...
        entry.metadata = metadata

    print(f"done. Please look at {qp.name} in QuPath. And look at he Project Metadata.")



from typing import TYPE_CHECKING

try:
    from ._version import version as __version__
except ImportError:  # pragma: no cover
    __version__ = "not-installed"

if TYPE_CHECKING:
    from dynaconf import Dynaconf
    settings: Dynaconf

__all__ = [
    "__version__",
    "settings",
]


def __getattr__(name):
    if name == "settings":
        from paquo._config import settings
        return settings
    else:
        raise AttributeError(name)



"""example showing how to draw a tile detection overlay over an image"""
import itertools
import math
import random
from pathlib import Path
from typing import Tuple, Iterator

from shapely.geometry import Polygon

from paquo.images import QuPathImageType
from paquo.projects import QuPathProject

EXAMPLE_PROJECT = Path(__file__).parent.absolute() / "projects" / "example_05_project"
EXAMPLE_IMAGE = Path(__file__).parent.absolute() / "images" / "image_1.svs"


def measurement(x, y, w, h, a=8) -> float:
    """some measurement that you want to display"""
    v = math.exp(-a * ((2 * x / w - 1) ** 2 + (2 * y / h - 1) ** 2))
    v += random.uniform(-0.1, 0.1)
    return min(max(0., v), 1.)


def iterate_grid(width, height, grid_size) -> Iterator[Tuple[int, int]]:
    """return corner x,y coordinates for a grid"""
    yield from itertools.product(
        range(0, width, grid_size),
        range(0, height, grid_size)
    )


with QuPathProject(EXAMPLE_PROJECT, mode='x') as qp:
    print("created", qp.name)
    # add an image
    entry = qp.add_image(
        EXAMPLE_IMAGE,
        image_type=QuPathImageType.BRIGHTFIELD_H_E
    )

    tile_size = 50
    img_width = entry.width
    img_height = entry.height

    # iterate over the image in a grid pattern
    for x0, y0 in iterate_grid(img_width, img_height, grid_size=tile_size):
        tile = Polygon.from_bounds(x0, y0, x0 + tile_size, y0 + tile_size)
        # add tiles (tiles are specialized detection objects drawn without border)
        detection = entry.hierarchy.add_tile(
            roi=tile,
            measurements={
                'measurement': measurement(x0, y0, img_width, img_height)
            }
        )

    print("added", len(entry.hierarchy.detections), "tiles")
    print(f"done. Please look at {qp.name} in QuPath and look at 'Measure > Show Measurement Maps'")



import pytest

from paquo.classes import QuPathPathClass


@pytest.fixture(scope='session')
def pathclass():
    yield QuPathPathClass("MyClass")


def test_pathclass_creation():
    with pytest.raises(TypeError):
        QuPathPathClass.from_java("abc")

    pc = QuPathPathClass("MyClass", color=None)
    assert pc.parent is None
    assert pc.name == pc.id == "MyClass"
    assert pc.is_valid
    assert "MyClass" in repr(pc)


def test_deny_name_none_creation():
    with pytest.raises(NotImplementedError):
        # noinspection PyTypeChecker
        QuPathPathClass(None, parent=None)

    pc = QuPathPathClass("MyClass")
    with pytest.raises(ValueError):
        # noinspection PyTypeChecker
        QuPathPathClass(None, parent=pc)


def test_incorrect_parent_type():
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        QuPathPathClass("new class", parent="parent_class")


def test_incorrect_class_name():
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        QuPathPathClass(1)
    with pytest.raises(ValueError):
        QuPathPathClass("my::class")


def test_pathclass_equality(pathclass):
    other = QuPathPathClass("MyClass2")
    same = QuPathPathClass("MyClass")
    assert pathclass == pathclass
    assert pathclass != other
    assert pathclass == same
    assert pathclass != 123


def test_pathclass_creation_with_parent(pathclass):
    pc = QuPathPathClass("MyChild", parent=pathclass)
    assert pc.parent == pathclass
    assert pc.name == "MyChild"
    assert pc.id == "MyClass: MyChild"

    assert pc.origin == pathclass
    assert pc.is_derived_from(pathclass)
    assert not pc.is_ancestor_of(pathclass)
    assert not pathclass.is_derived_from(pc)
    assert pathclass.is_ancestor_of(pc)
    assert pc.is_derived_class
    assert not pathclass.is_derived_class


def test_pathclass_colors():
    pc = QuPathPathClass("MyNew", color=None)
    my_class_color = (49, 139, 153)  # based on string MyNew
    assert pc.color.to_rgb() == my_class_color

    pc = QuPathPathClass("MyNew2", color=(1, 2, 3))
    assert pc.color.to_rgb() == (1, 2, 3)

    pc.color = "#ff0000"
    assert pc.color.to_rgb() == (255, 0, 0)


def test_pathclass_none_colors():
    pc = QuPathPathClass("MyNew")
    pc.color = None
    assert pc.color is None



import collections
import json
import math
import reprlib
import struct
import warnings
from contextlib import contextmanager
from contextlib import suppress
from typing import Any
from typing import Counter as CounterType
from typing import Iterable
from typing import Iterator
from typing import MutableSet
from typing import Optional
from typing import Sequence
from typing import Type
from typing import Union
from typing import overload

from paquo._logging import get_logger
from paquo._utils import cached_property
from paquo.classes import QuPathPathClass
from paquo.java import GsonTools
from paquo.java import IllegalArgumentException
from paquo.java import PathAnnotationObject
from paquo.java import PathDetectionObject
from paquo.java import PathObjectHierarchy
from paquo.java import PathTileObject
from paquo.java import String
from paquo.java import compatibility
from paquo.pathobjects import BaseGeometry
from paquo.pathobjects import PathROIObjectType
from paquo.pathobjects import QuPathPathAnnotationObject
from paquo.pathobjects import QuPathPathCellObject
from paquo.pathobjects import QuPathPathDetectionObject
from paquo.pathobjects import QuPathPathTileObject
from paquo.pathobjects import fix_geojson_geometry

__all__ = ["QuPathPathObjectHierarchy"]

_logger = get_logger(__name__)


class PathObjectProxy(Sequence[PathROIObjectType], MutableSet[PathROIObjectType]):
    """set interface for path objects with support for access by index and slicing

    *not meant to be instantiated by the user*

    Notes
    -----
    Access this proxy via the `QuPathPathObjectHierarchy.annotations` or
    `QuPathPathObjectHierarchy.detections` properties. It acts just like
    a python set, but also supports access by index and slicing.

    """

    def __init__(
        self,
        hierarchy: 'QuPathPathObjectHierarchy',
        paquo_cls: Type[PathROIObjectType],
        mask: Optional[Union[slice, Sequence[int]]] = None,
        readonly: Optional[bool] = None,
    ) -> None:
        """internal: not meant to be instantiated by the user"""
        self._hierarchy = hierarchy
        self._paquo_cls = paquo_cls
        if not (
            mask is None
            or isinstance(mask, slice)
            or (all(isinstance(x, int) for x in mask) and len(mask) > 0)
        ):
            raise TypeError(f"mask can be slice, or Sequence[int] or None. Got: {type(mask)!r}")
        self._mask: Optional[Union[slice, Sequence[int]]] = mask
        self._init_readonly = readonly

    @property
    def _readonly(self) -> bool:
        if self._init_readonly is not None:
            return self._init_readonly
        # noinspection PyProtectedMember
        return self._hierarchy._readonly or self._mask is not None

    @property
    def _java_hierarchy(self):
        return self._hierarchy.java_object

    @cached_property
    def _list(self):
        _list = self._java_hierarchy.getObjects(None, self._paquo_cls.java_class)
        if self._mask:
            if isinstance(self._mask, slice):
                _list = _list[self._mask]
            else:
                _list = [_list[x] for x in self._mask]
        return _list

    def _list_invalidate_cache(self):
        with suppress(KeyError):
            del self.__dict__["_list"]

    def _disabled(self, other: Iterable[Any]) -> "PathObjectProxy":
        raise NotImplementedError(f"{type(self).__name__} only supports inplace operations: '|=', '-='")
    __or__ = __and__ = __sub__ = __xor__ = __iand__ = __ixor__ = _disabled
    del _disabled

    def __ior__(self, other: Iterable[Any]) -> "PathObjectProxy":  # type: ignore
        if self._mask:
            raise OSError("cannot modify view")
        if self._readonly:
            raise OSError("project in readonly mode")
        path_objects = [x.java_object for x in other]
        try:
            self._java_hierarchy.addPathObjects(path_objects)
        finally:
            self._list_invalidate_cache()
        return self
    update = __ior__

    def __isub__(self, other: Iterable[Any]) -> "PathObjectProxy":  # type: ignore
        if self._mask:
            raise OSError("cannot modify view")
        if self._readonly:
            raise OSError("project in readonly mode")
        path_objects = [x.java_object for x in other]
        try:
            self._java_hierarchy.removeObjects(path_objects, True)
        finally:
            self._list_invalidate_cache()
        return self

    if compatibility.supports_newer_addobject_and_pathclass():
        def add(self, x: PathROIObjectType) -> None:
            """adds a new path object to the proxy"""
            if self._mask:
                raise OSError("cannot modify view")
            if self._readonly:
                raise OSError("project in readonly mode")
            if not isinstance(x, self._paquo_cls):
                raise TypeError(f"requires {self._paquo_cls.__name__} instance got {x.__class__.__name__}")
            try:
                if self._hierarchy.autoflush:
                    self._java_hierarchy.addObject(x.java_object, True)
                else:
                    self._java_hierarchy.addObject(x.java_object, False)
            finally:
                self._list_invalidate_cache()

    else:
        def add(self, x: PathROIObjectType) -> None:
            """adds a new path object to the proxy"""
            if self._mask:
                raise OSError("cannot modify view")
            if self._readonly:
                raise OSError("project in readonly mode")
            if not isinstance(x, self._paquo_cls):
                raise TypeError(f"requires {self._paquo_cls.__name__} instance got {x.__class__.__name__}")
            try:
                if self._hierarchy.autoflush:
                    self._java_hierarchy.addPathObject(x.java_object)
                else:
                    self._java_hierarchy.addPathObjectWithoutUpdate(x.java_object)
            finally:
                self._list_invalidate_cache()

    def discard(self, x: PathROIObjectType) -> None:
        """discard a path object from the proxy"""
        if self._mask:
            raise OSError("cannot modify view")
        if self._readonly:
            raise OSError("project in readonly mode")
        if not isinstance(x, self._paquo_cls):
            raise TypeError(f"requires {self._paquo_cls.__name__} instance got {x.__class__.__name__}")
        try:
            if self._hierarchy.autoflush:
                self._java_hierarchy.removeObject(x.java_object, True)
            else:
                self._java_hierarchy.removeObjectWithoutUpdate(x.java_object, True)
        finally:
            self._list_invalidate_cache()

    def clear(self) -> None:
        """clear all path objects from the proxy"""
        if self._mask:
            raise OSError("cannot modify view")
        if self._readonly:
            raise OSError("project in readonly mode")
        try:
            self._java_hierarchy.removeObjects(self._list, True)
        finally:
            self._list_invalidate_cache()

    def __contains__(self, x: Any) -> bool:
        """test if path object is in proxy"""
        # ... inHierarchy is private
        # return bool(self._java_hierarchy.inHierarchy(x.java_object))
        if not isinstance(x, self._paquo_cls):
            return False
        while x.parent is not None:
            x = x.parent
        return bool(x.java_object == self._java_hierarchy.getRootObject())

    def __len__(self) -> int:
        return len(self._list)

    def __iter__(self: "PathObjectProxy") -> Iterator[PathROIObjectType]:
        for obj in self._list:
            yield self._paquo_cls(obj, update_callback=self.add)

    @overload
    def __getitem__(self, i: int) -> PathROIObjectType: ...
    @overload
    def __getitem__(self, i: slice) -> "PathObjectProxy": ...
    @overload
    def __getitem__(self, i: Sequence[int]) -> "PathObjectProxy": ...

    def __getitem__(self, i):
        if isinstance(i, int):
            return self._paquo_cls(self._list[i], update_callback=self.add)
        elif isinstance(i, slice):
            if self._mask is None:
                mask = i
            elif isinstance(self._mask, slice):
                _r = range(self._java_hierarchy.nObjects())
                _s = _r[self._mask][i]
                mask = slice(_s.start, _s.stop, _s.step)
            else:
                mask = self._mask[i]
            return PathObjectProxy(self._hierarchy, self._paquo_cls, mask)
        else:
            if self._mask is None:
                mask = i
            elif isinstance(self._mask, slice):
                _r = range(self._java_hierarchy.nObjects())
                _s = _r[self._mask]
                mask = [_s[idx] for idx in i]
            else:
                mask = [self._mask[idx] for idx in i]
            return PathObjectProxy(self._hierarchy, self._paquo_cls, mask)

    def count(self, value: PathROIObjectType) -> int:
        return int(value in self)  # PathObjectProxy is a set

    def __repr__(self):
        c = type(self).__name__
        h = repr(self._hierarchy)
        p = self._paquo_cls.__name__
        m = reprlib.repr(self._mask)
        i = f"0x{hex(id(self))}"
        if m is None:
            return f"<{c} hierarchy={h} paquo_cls={p} at {i}>"
        return f"<{c} hierarchy={h} paquo_cls={p} mask={m} at {i}>"


class QuPathPathObjectHierarchy:
    java_object: PathObjectHierarchy

    def __init__(
        self,
        hierarchy: Optional[PathObjectHierarchy] = None,
        *,
        readonly: bool = False,
        image_name: str = "N/A",
        autoflush: bool = True,
    ) -> None:
        """qupath hierarchy stores all annotation objects

        Parameters
        ----------
        hierarchy:
            a PathObjectHierarchy instance (optional)
            Usually accessed directly via the Image Container.
        """
        if hierarchy is None:
            hierarchy = PathObjectHierarchy()
        self.java_object = hierarchy
        # internals
        self._image_name = str(image_name)
        self._readonly = bool(readonly)
        self._annotations = PathObjectProxy(self, paquo_cls=QuPathPathAnnotationObject)
        self._detections = PathObjectProxy(self, paquo_cls=QuPathPathDetectionObject)
        self._cells = PathObjectProxy(self, paquo_cls=QuPathPathCellObject, readonly=True)
        self._tiles = PathObjectProxy(self, paquo_cls=QuPathPathTileObject, readonly=True)
        # attrs
        self.autoflush = bool(autoflush)

    def __len__(self) -> int:
        """Number of objects in hierarchy (all types)"""
        return int(self.java_object.nObjects())

    @property
    def is_empty(self) -> bool:
        """a hierarchy is empty if it only contains the root object"""
        return bool(self.java_object.isEmpty())

    @contextmanager
    def no_autoflush(self):
        """prevent updates to the hierarchy to trigger an internal update event"""
        _autoflush, self.autoflush = self.autoflush, False
        try:
            yield self
        finally:
            self.flush(invalidate_proxy_cache=True)
            self.autoflush = _autoflush

    # noinspection PyProtectedMember
    def flush(self, invalidate_proxy_cache: bool = False):
        """flush changes to the hierarchy"""
        root = self.java_object.getRootObject()
        self.java_object.fireHierarchyChangedEvent(root)
        if invalidate_proxy_cache:
            self._annotations._list_invalidate_cache()
            self._detections._list_invalidate_cache()
            self._cells._list_invalidate_cache()
            self._tiles._list_invalidate_cache()

    @property
    def root(self) -> QuPathPathAnnotationObject:
        """the hierarchy root node

        This object has no roi and cannot be assigned another class.
        All other objects are descendants of this object if they are
        attached to this hierarchy.
        """
        root = self.java_object.getRootObject()
        return QuPathPathAnnotationObject(root)  # todo: specialize...

    @property
    def annotations(self) -> PathObjectProxy[QuPathPathAnnotationObject]:
        """all annotations provided as a flattened set-like proxy"""
        return self._annotations

    def add_annotation(self,
                       roi: BaseGeometry,
                       path_class: Optional[QuPathPathClass] = None,
                       measurements: Optional[dict] = None,
                       *,
                       path_class_probability: float = math.nan) -> QuPathPathAnnotationObject:
        """convenience method for adding annotations"""
        if self._readonly:
            raise OSError("project in readonly mode")
        obj = QuPathPathAnnotationObject.from_shapely(
            roi, path_class, measurements,
            path_class_probability=path_class_probability
        )
        self._annotations.add(obj)
        return obj

    @property
    def detections(self) -> PathObjectProxy[QuPathPathDetectionObject]:
        """all detections provided as a flattened set-like proxy"""
        return self._detections

    def add_detection(self,
                      roi: BaseGeometry,
                      path_class: Optional[QuPathPathClass] = None,
                      measurements: Optional[dict] = None,
                      *,
                      path_class_probability: float = math.nan) -> QuPathPathDetectionObject:
        if self._readonly:
            raise OSError("project in readonly mode")
        """convenience method for adding detections

        Notes
        -----
        these will be added to self.detections
        """
        obj = QuPathPathDetectionObject.from_shapely(
            roi, path_class, measurements,
            path_class_probability=path_class_probability
        )
        self._detections.add(obj)
        return obj

    @property
    def tiles(self) -> PathObjectProxy[QuPathPathTileObject]:
        """all tiles provided as a flattened read-only set-like proxy"""
        return self._tiles

    def add_tile(self,
                 roi: BaseGeometry,
                 path_class: Optional[QuPathPathClass] = None,
                 measurements: Optional[dict] = None,
                 *,
                 path_class_probability: float = math.nan) -> QuPathPathTileObject:
        """convenience method for adding tile detections

        Notes
        -----
        these will be added to self.detections
        """
        if self._readonly:
            raise OSError("project in readonly mode")
        obj = QuPathPathTileObject.from_shapely(
            roi, path_class, measurements,
            path_class_probability=path_class_probability
        )
        self._detections.add(obj)
        return obj

    @property
    def cells(self) -> PathObjectProxy[QuPathPathCellObject]:
        """all cells provided as a flattened read-only set-like proxy"""
        return self._cells

    def add_cell(self,
                 roi: BaseGeometry,
                 path_class: Optional[QuPathPathClass] = None,
                 measurements: Optional[dict] = None,
                 *,
                 path_class_probability: float = math.nan,
                 nucleus_roi: BaseGeometry) -> QuPathPathCellObject:
        """convenience method for adding cell detections

        Notes
        -----
        these will be added to self.detections, as they are a subclass of detections
        """
        if self._readonly:
            raise OSError("project in readonly mode")
        obj = QuPathPathCellObject.from_shapely(
            roi, path_class, measurements,
            path_class_probability=path_class_probability,
            nucleus_roi=nucleus_roi
        )
        self._detections.add(obj)
        return obj

    def to_geojson(self) -> list:
        """return all annotations as a list of geojson features"""
        gson = GsonTools.getInstance()
        geojson = gson.toJson(self.java_object.getAnnotationObjects())
        return list(json.loads(str(geojson)))

    def load_geojson(
            self, geojson: list,
            *, raise_on_skip: bool = False, fix_invalid: bool = False,
    ) -> bool:
        """load annotations into this hierarchy from a geojson list

        returns True if new objects were added, False otherwise.
        """
        # todo: use geojson module for type checking?
        if self._readonly:
            raise OSError("project in readonly mode")
        if not isinstance(geojson, list):
            raise TypeError("requires a geojson list")

        requires_annotation_json_fix = compatibility.requires_annotation_json_fix()

        aos = []
        skipped: "CounterType[str]" = collections.Counter()
        for annotation in geojson:
            try:
                if fix_invalid:
                    annotation["geometry"] = fix_geojson_geometry(annotation["geometry"])

                properties = annotation["properties"]
                if "objectType" in properties:
                    # https://github.com/qupath/qupath/pull/1099
                    object_type = properties["objectType"]
                elif "object_type" in properties:
                    object_type = properties["object_type"]
                else:
                    object_type = "unknown"

                # compatibility layer
                if (
                    requires_annotation_json_fix
                    and 'id' not in annotation
                ):
                    object_id = {
                        'annotation': "PathAnnotationObject",
                        'detection': "PathDetectionObject",
                        'tile': "PathTileObject",
                        'cell': "PathCellObject",
                        'tma_core': "TMACoreObject",
                        'tmaCore': "TMACoreObject",  # https://github.com/qupath/qupath/pull/1099
                        'root': "PathRootObject",
                        'unknown': "PathAnnotationObject",
                    }.get(object_type, None)
                    if object_id is None:
                        _logger.warn(f"annotation has incompatible object_type: '{object_type}'")
                        object_id = "PathAnnotationObject"
                    annotation['id'] = object_id

                gson = GsonTools.getInstance()
                if object_type == "annotation":
                    java_obj = gson.fromJson(String(json.dumps(annotation)), PathAnnotationObject)
                elif object_type == "detection":
                    java_obj = gson.fromJson(String(json.dumps(annotation)), PathDetectionObject)
                elif object_type == "tile":
                    java_obj = gson.fromJson(String(json.dumps(annotation)), PathTileObject)
                else:
                    if object_type != "unknown":
                        warnings.warn(
                            f"Trying to load annotation object_type={object_type!r}. "
                            "Please report this on the paquo issue tracker on github and provide a geojson example. "
                            "https://github.com/Bayer-Group/paquo/issues",
                            stacklevel=2,
                        )
                    java_obj = gson.fromJson(String(json.dumps(annotation)), PathAnnotationObject)

            except (IllegalArgumentException, ValueError) as err:
                _logger.warn(f"Annotation skipped: {err}")
                class_ = annotation["properties"].get("classification", {}).get("name", "UNDEFINED")
                skipped[class_] += 1
                continue

            else:
                aos.append(java_obj)

        if skipped:
            n_skipped = sum(skipped.values())
            if raise_on_skip:
                raise ValueError(f"could not convert {n_skipped} annotations")
            _logger.error(
                f"skipped {n_skipped} annotation objects: {skipped.most_common()}"
            )

        updated = bool(self.java_object.insertPathObjects(aos))
        if updated:
            self.flush(invalidate_proxy_cache=True)
        return updated

    def to_ome_xml(self, prefix="paquo", fill_alpha=0.0) -> str:
        """return all annotations in ome xml format"""
        # this
        try:
            from ome_types import to_xml
            from ome_types.model import OME
            from ome_types.model import ROI
            from ome_types.model import AnnotationRef
            from ome_types.model import Ellipse as OmeEllipse
            from ome_types.model import Line as OmeLine
            from ome_types.model import Map
            from ome_types.model import MapAnnotation
            from ome_types.model import Point as OmePoint
            from ome_types.model import Polygon as OmePolygon
            from ome_types.model import Polyline as OmePolyline
            from ome_types.model import Rectangle as OmeRectangle
            from ome_types.model.map import M
            from ome_types.model.shape import FillRule
        except ImportError:
            raise RuntimeError(f"{type(self).__name__}.to_ome_xml requires 'ome-types' python module and python>=3.7")

        from paquo.java import EllipseROI
        from paquo.java import GeometryROI
        from paquo.java import LineROI
        from paquo.java import PointsROI
        from paquo.java import PolygonROI
        from paquo.java import PolylineROI
        from paquo.java import RectangleROI

        ome = OME()

        for ao in self.annotations:

            class_name: Optional[str]
            if ao.path_class:
                class_name = ao.path_class.name
            else:
                class_name = None

            # --- create the map_annotation

            _m = {
                f"{prefix}:object_type": {
                    "PathAnnotationObject": 'annotation',
                    "PathDetectionObject": 'detection',
                    "PathTileObject": 'tile',
                    "PathCellObject": 'cell',
                    "TMACoreObject": 'tma_core',
                    "PathRootObject": 'root',
                }.get(type(ao.java_object).__name__.rpartition(".")[2], "unknown")
            }
            if class_name:
                _m[f"{prefix}:path_class"] = class_name
            if ao.name:
                _m[f"{prefix}:name"] = ao.name
            for k, v in ao.measurements.items():
                _m[f"{prefix}:measurement:{k}"] = v

            if "ms" in Map.__annotations__:
                map_annotation = MapAnnotation(  # type: ignore
                    value=Map(
                        ms=[
                            M(k=_key, value=str(_value))
                            for _key, _value in _m.items()
                        ]
                    )
                )
            else:
                # ome-types < 0.4.0
                map_annotation = MapAnnotation(  # type: ignore
                    value=Map(
                        m=[
                            M(k=_key, value=str(_value))
                            for _key, _value in _m.items()
                        ]
                    )
                )

            # --- prepare common kwargs for ome Shape

            if ao.path_class and ao.path_class.color:
                # https://www.openmicroscopy.org/Schemas/Documentation/Generated/OME-2016-06/ome_xsd.html#Color
                r, g, b, a = ao.path_class.color.to_rgba()
                stroke_color, = struct.unpack(">i", bytes([r, g, b, a]))
                if fill_alpha <= 0:
                    fill_color = None
                else:
                    fill_alpha = min(max(0.0, fill_alpha), 1.0)
                    fill_color, = struct.unpack(">i", bytes([r, g, b, int(fill_alpha*a)]))
            else:
                stroke_color = None
                fill_color = None
            # https://www.openmicroscopy.org/Schemas/Documentation/Generated/OME-2016-06/ome_xsd.html#Shape_FillRule
            fill_rule = FillRule.NON_ZERO

            qp_roi = ao.java_object.getROI()
            the_c = int(qp_roi.getC())
            the_t = int(qp_roi.getT())
            the_z = int(qp_roi.getZ())

            shape_kwargs = dict(
                # id=...,
                # annotation_ref=...,
                fill_color=fill_color,
                fill_rule=fill_rule,
                # font_family=None,
                # font_size=None,
                # font_size_unit=UnitsLength("pt"),
                # font_style=None,
                locked=ao.locked,
                stroke_color=stroke_color,
                stroke_dash_array=None,
                stroke_width=2,  # 2px is the QuPath default gui setting
                # stroke_width_unit=UnitsLength("pixel"),
                text=class_name,  # again class instead of ao.description for a nicer UX in omero.iviewer...
                the_c=the_c if the_c >= 0 else None,
                the_t=the_t,
                the_z=the_z,
                transform=None,
            )

            # --- create the roi

            roi = ROI(name=class_name)  # type: ignore
            try:
                roi.annotation_refs.append(AnnotationRef(id=map_annotation.id))  # type: ignore
            except AttributeError:
                # ome-types<0.4.0
                roi.annotation_ref.append(AnnotationRef(id=map_annotation.id))

            # --- add the correct shape dependent on roi types

            ome_shape: Any
            if isinstance(qp_roi, (PolygonROI, GeometryROI)):
                ome_shape = OmePolygon(
                    points=" ".join(f"{p.getX():f},{p.getY():f}" for p in qp_roi.getAllPoints()),
                    **shape_kwargs,
                )
            elif isinstance(qp_roi, EllipseROI):
                # https://github.com/qupath/qupath/blob/e84467e86751e5aa542ab68a4915b70ecbf2f6fc/qupath-core/src/main/java/qupath/lib/roi/EllipseROI.java#L60-L63
                ome_shape = OmeEllipse(
                    radius_x=float(qp_roi.getBoundsWidth() * 0.5),
                    radius_y=float(qp_roi.getBoundsHeight() * 0.5),
                    x=float(qp_roi.getCentroidX()),
                    y=float(qp_roi.getCentroidY()),
                    **shape_kwargs,
                )
            elif isinstance(qp_roi, RectangleROI):
                ome_shape = OmeRectangle(
                    height=float(qp_roi.getBoundsHeight()),
                    width=float(qp_roi.getBoundsWidth()),
                    x=float(qp_roi.getBoundsX()),
                    y=float(qp_roi.getBoundsY()),
                    **shape_kwargs,
                )
            elif isinstance(qp_roi, LineROI):
                ome_shape = OmeLine(
                    x1=float(qp_roi.getX1()),
                    x2=float(qp_roi.getX2()),
                    y1=float(qp_roi.getY1()),
                    y2=float(qp_roi.getY2()),
                    marker_end=None,
                    marker_start=None,
                    **shape_kwargs,
                )
            elif isinstance(qp_roi, PolylineROI):
                ome_shape = OmePolyline(
                    points=" ".join(f"{p.getX():f},{p.getY():f}" for p in qp_roi.getAllPoints()),
                    marker_end=None,
                    marker_start=None,
                    **shape_kwargs,
                )
            elif isinstance(qp_roi, PointsROI):
                # we have to create individual points in ome
                ome_shape = [
                    OmePoint(x=float(p.getX()), y=float(p.getY()), **shape_kwargs)
                    for p in qp_roi.getAllPoints()
                ]
            else:
                raise NotImplementedError(f"todo {type(qp_roi).__name__}")

            if isinstance(ome_shape, list):
                roi.union.extend(ome_shape)
            else:
                roi.union.append(ome_shape)

            # --- add the annotation to the ome structure
            ome.rois.append(roi)
            ome.structured_annotations.append(map_annotation)  # type: ignore[union-attr]

        return to_xml(ome)

    def __repr__(self):
        return f"Hierarchy(image={self._image_name}, annotations={len(self._annotations)}, detections={len(self._detections)})"

    def _repr_html_(self):
        from paquo._repr import br
        from paquo._repr import div
        from paquo._repr import h4
        from paquo._repr import p
        from paquo._repr import span

        return div(
            h4(text=f"Hierarchy: {self._image_name}", style={"margin-top": "0"}),
            p(
                span(text="annotations: ", style={"font-weight": "bold"}),
                span(text=f"{len(self._annotations)}"),
                br(),
                span(text="detections: ", style={"font-weight": "bold"}),
                span(text=f"{len(self._detections)}"),
                style={"margin": "0.5em"},
            ),
        )



import collections.abc as collections_abc
import math
import pathlib
import re
import shutil
from contextlib import contextmanager
from contextlib import nullcontext
from typing import Any
from typing import Callable
from typing import ContextManager
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Literal
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union
from typing import overload

from paquo import settings
from paquo._logging import get_logger
from paquo._logging import redirect
from paquo._utils import make_backup_filename
from paquo.classes import QuPathPathClass
from paquo.images import ImageProvider
from paquo.images import QuPathImageType
from paquo.images import QuPathProjectImageEntry
from paquo.images import SimpleFileImageId
from paquo.java import URI
from paquo.java import BufferedImage
from paquo.java import DefaultProject
from paquo.java import ExceptionInInitializerError
from paquo.java import File
from paquo.java import Files
from paquo.java import GeneralTools
from paquo.java import ImageServerProvider
from paquo.java import IOException
from paquo.java import NegativeArraySizeException
from paquo.java import ProjectImportImagesCommand_getThumbnailRGB
from paquo.java import ProjectIO
from paquo.java import Projects
from paquo.java import ServerTools
from paquo.java import String
from paquo.java import compatibility

_log = get_logger(__name__)


class _ProjectImageEntriesProxy(collections_abc.Sequence):
    """iterable container holding image entries"""
    # todo: decide if this should be a mapping or not...
    #   maybe with key id? to simplify re-association

    def __init__(self, project: 'QuPathProject'):
        if not isinstance(project.java_object, DefaultProject):
            raise TypeError('requires DefaultProject instance')  # pragma: no cover
        self._project = project
        self._images = {
            self._key_func(entry): QuPathProjectImageEntry(entry, _project_ref=self._project)
            for entry in self._project.java_object.getImageList()
        }

    def _key_func(self, entry):
        """retrieve the fullProjectID from an ImageEntry

        note: this is only valid for the current project instance
        """
        # basically `DefaultProjectImageEntry.getFullProjectEntryID()`
        # but don't go via image_data
        return (
            str(self._project.java_object.getPath().toAbsolutePath().toString()),
            str(entry.getID()),
        )

    def refresh(self):
        removed = set(self._images.keys())
        for entry in self._project.java_object.getImageList():
            key = self._key_func(entry)
            if key not in self._images:
                self._images[key] = QuPathProjectImageEntry(entry, _project_ref=self._project)
            else:
                removed.discard(key)  # existing entry
        if removed:
            for key in removed:  # pragma: no cover
                _ = self._images.pop(key)

    def __iter__(self) -> Iterator[QuPathProjectImageEntry]:
        return iter(self._images.values())

    def __contains__(self, entry: object) -> bool:
        if not isinstance(entry, QuPathProjectImageEntry):
            return False
        full_id = self._key_func(entry.java_object)
        return full_id in self._images

    def __len__(self) -> int:
        return len(self._images)

    def __repr__(self):
        return f"ImageEntries({repr([entry.image_name for entry in self])})"

    def _repr_html_(self, compact=False) -> str:
        from paquo._repr import div
        from paquo._repr import h4
        from paquo._repr import repr_html
        images = [
            repr_html(img, compact=True, index=idx) for idx, img in enumerate(self)
        ]
        header_css = {"margin-top": "0"} if not compact else {}
        return div(
            h4(text="Images:", style=header_css),
            div(*images, style={"display": "flex"}),
        )

    @overload
    def __getitem__(self, i: int) -> QuPathProjectImageEntry: ...

    @overload
    def __getitem__(self, i: slice) -> Sequence[QuPathProjectImageEntry]: ...

    def __getitem__(self, i):
        if not isinstance(i, (int, slice)):
            raise IndexError(i)
        return list(self._images.values())[i]


def _stash_project_files(project_dir: pathlib.Path):
    """move rename projects files in a project to .backup"""
    if not project_dir.is_dir():
        return
    if not any(project_dir.iterdir()):
        return

    if settings.safe_truncate:
        # create a backup archive of the project
        _log.warning(f"backing up {project_dir.name} before truncating!"
                     " (this can be disabled via PAQUO_SAFE_TRUNCATE)")
        str_name = shutil.make_archive(
            f".{project_dir.name}",
            format="zip",
            root_dir=project_dir.parent,
            base_dir=project_dir,
            logger=_log
        )
        # and rename it
        src_name = pathlib.Path(str_name)
        dst_name = make_backup_filename(project_dir.parent, src_name.stem)
        shutil.move(str(src_name), str(dst_name))
        _log.info(f"backup to {dst_name.name} successful")

    # empty the project directory
    for p in project_dir.iterdir():
        if p.is_file():
            p.unlink()
        elif p.is_dir():
            shutil.rmtree(p)
    # done


DEFAULT_IMAGE_PROVIDER: Any = ImageProvider()


ProjectIOMode = Literal["r", "r+", "w", "w+", "a", "a+", "x", "x+"]


class QuPathProject:
    java_object: DefaultProject

    def __init__(self,
                 path: Union[str, pathlib.Path],
                 mode: ProjectIOMode = 'r',
                 *,
                 image_provider: ImageProvider = DEFAULT_IMAGE_PROVIDER):
        """load or create a new qupath project

        Parameters
        ----------
        path:
            path to `project.qpproj` file, or its parent directory
        mode:
            'r' --> readonly, error if not there
            'r+' --> read/write, error if not there
            'a' = 'a+' --> read/write, create if not there, append if there
            'w' = 'w+' --> read/write, create if not there, truncate if there
            'x' = 'x+' --> read/write, create if not there, error if there

        """
        if not isinstance(image_provider, ImageProvider):
            raise TypeError("image_provider must quack like a paquo.images.ImageProvider")

        self._path = pathlib.Path(path)
        self._mode = str(mode)

        # guarantee p points to qpproj file (allow directory)
        if not self._path.suffix:
            self._path /= 'project.qpproj'
        elif self._path.suffix != ".qpproj":
            raise ValueError("project file requires '.qpproj' suffix")

        if not re.match(r"^[rawx][+]?$", mode):
            raise ValueError(f"unsupported mode '{mode}'")

        p = self._path.expanduser().absolute()
        _exists = p.is_file()
        self._readonly = mode == "r"

        if mode in {"r", "r+"} and not _exists:
            raise FileNotFoundError(p)
        elif mode in {"x", "x+"} and _exists:
            raise FileExistsError(p)
        elif mode in {"w", "w+"}:
            # truncate gracefully
            p_dir = p.parent
            if p_dir.is_dir():
                _stash_project_files(p_dir)
            _exists = False

        if _exists:
            cm: Callable[..., ContextManager[Any]]
            if compatibility.requires_missing_classes_json_fix():
                @contextmanager
                def cm(is_readonly):
                    classes_json = p.parent.joinpath("classifiers", "classes.json")
                    is_missing = not classes_json.is_file()
                    if is_missing:
                        classes_json.write_text('{"pathClasses":[]}')
                    try:
                        yield
                    finally:
                        if is_missing and is_readonly:
                            classes_json.unlink(missing_ok=True)
            else:
                cm = nullcontext

            with redirect(stderr=True, stdout=True), cm(self._readonly):
                project = ProjectIO.loadProject(File(str(p)), BufferedImage)
        else:
            p_dir = p.parent
            p_dir.mkdir(parents=True, exist_ok=True)
            for f in p_dir.iterdir():
                if not f.match("*.backup"):
                    raise ValueError("will only create projects in empty directories")
            with redirect(stderr=True, stdout=True):
                project = Projects.createProject(File(str(p_dir)), BufferedImage)

        self.java_object = project
        self._image_entries_proxy = _ProjectImageEntriesProxy(self)
        self._image_provider = image_provider

    @property
    def images(self) -> Sequence[QuPathProjectImageEntry]:
        """project images"""
        return self._image_entries_proxy

    @contextmanager
    def _stage_image_entry(self, server_builder):
        """internal contextmanager for staging new entries"""
        entry = self.java_object.addImage(server_builder)
        try:
            yield entry
        except Exception:
            # todo: check if we could set removeAllData to True here
            self.java_object.removeImage(entry, False)
            raise
        finally:
            # update the proxy
            self._image_entries_proxy.refresh()

    @overload
    def add_image(
        self,
        image_id: SimpleFileImageId,
        image_type: Optional[QuPathImageType] = ...,
        *,
        allow_duplicates: bool = ...,
        return_list: Literal[True],
    ) -> List[QuPathProjectImageEntry]:
        ...

    @overload
    def add_image(
        self,
        image_id: SimpleFileImageId,
        image_type: Optional[QuPathImageType] = ...,
        *,
        allow_duplicates: bool = ...,
        return_list: Literal[False] = ...,
    ) -> Union[QuPathProjectImageEntry, List[QuPathProjectImageEntry]]:
        ...

    @redirect(stderr=True, stdout=True)
    def add_image(
        self,
        image_id: SimpleFileImageId,
        image_type: Optional[QuPathImageType] = None,
        *,
        allow_duplicates: bool = False,
        return_list: bool = False,
    ) -> Union[QuPathProjectImageEntry, List[QuPathProjectImageEntry]]:
        """add an image to the project

        Parameters
        ----------
        image_id:
            image_id pointing to the image file (with default image_provider: filename)
        image_type:
            provide an image type for the image. If not provided the user will
            be prompted before opening the image in QuPath.
        allow_duplicates:
            check if file has already been added to the project.

        """
        # readonly?
        if self._readonly:
            raise OSError("project in readonly mode")
        # test if we may add:
        img_uri = self._image_provider.uri(image_id)
        if img_uri is None:
            raise FileNotFoundError(f"image_provider can't provide URI for requested image_id: '{image_id}'")
        # img_id = self._image_provider.id(img_uri)
        # if img_id != image_id:  # pragma: no cover
        #     _log.warning(f"image_provider roundtrip error: '{image_id}' -> uri -> '{img_id}'")

        if not allow_duplicates:
            for entry in self.images:
                if img_uri == self._image_provider.uri(entry.uri):
                    raise FileExistsError(image_id)

        # first get a server builder
        try:
            support = ImageServerProvider.getPreferredUriImageSupport(
                BufferedImage,
                String(img_uri),
            )
        except IOException:  # pragma: no cover
            # it's possible that an image_provider returns an URI but that URI
            # is not actually reachable. In that case catch the java IOException
            # and raise a FileNotFoundError here
            raise FileNotFoundError(f"{image_id!r} as {img_uri!r}")
        except ExceptionInInitializerError:
            raise OSError("no preferred support found")
        if not support:
            raise OSError("no preferred support found")  # pragma: no cover
        server_builders = list(support.getBuilders())
        if not server_builders:
            raise OSError("no supported server builders found")  # pragma: no cover

        entries = []
        for server_builder in server_builders:
            with self._stage_image_entry(server_builder) as j_entry:
                # all of this happens in qupath.lib.gui.commands.ProjectImportImagesCommand
                try:
                    server = server_builder.build()
                except IOException:
                    _, _, _sb = server_builder.__class__.__name__.rpartition(".")
                    raise OSError(f"{_sb} can't open {str(image_id)}")
                j_entry.setImageName(ServerTools.getDisplayableImageName(server))

                # add some informative logging
                _md = server.getMetadata()
                width = int(_md.getWidth())
                height = int(_md.getHeight())
                downsamples = [float(x) for x in _md.getPreferredDownsamplesArray()]
                target_downsample = math.sqrt(width / 1024.0 * height / 1024.0)
                _log.info(f"Image[{width}x{height}] with downsamples {downsamples}")
                if not any(d >= target_downsample for d in downsamples):
                    _log.warning("No matching downsample for thumbnail! This might take a long time...")

                # set the project thumbnail
                try:
                    thumbnail = ProjectImportImagesCommand_getThumbnailRGB(server, None)
                except NegativeArraySizeException:  # pragma: no cover
                    raise RuntimeError(
                        "Thumbnailing FAILED. Image might be too large and has no embedded thumbnail."
                    )
                else:
                    j_entry.setThumbnail(thumbnail)

            py_entry = self._image_entries_proxy[-1]
            if image_type is not None:
                py_entry.image_type = image_type
            # save project after adding image
            py_entry.save()
            self.save(images=False)
            entries.append(py_entry)
        if return_list or len(entries) > 1:
            return entries
        else:
            return entries[0]

    def is_readable(self) -> Dict[str, bool]:
        """verify if images are reachable"""
        readability_map = {}
        for image in self.images:
            uri = image.uri
            if uri is None:
                raise RuntimeError(f"entry has None uri: {image!r}")
            if uri in readability_map:  # pragma: no cover
                raise RuntimeError("received the same image_id from image_provider for two different images")
            readability_map[str(uri)] = image.is_readable()
        return readability_map

    def update_image_paths(self, *, try_relative: bool = False, **rebase_kwargs) -> None:
        """update image path uris if image files moved

        Parameters
        ----------
        try_relative: bool
            if try_relative is True, update_image_paths tries to use relative paths
            to resolve missing image entries. This is useful in case you move the
            the project together with its images directory, or in case you remount
            at a different location.
        **rebase_kwargs:
            keyword arguments are handed over to the image provider instance.
            The default image provider is a paquo.images.ImageProvider
            which uses the uri2uri keyword argument. (A mapping from old URI to new
            URI: Mapping[str, str])

        """
        if not isinstance(try_relative, bool):
            raise TypeError(f"try_relative requires bool, got: {type(try_relative).__name__!r}")

        # get the current image URIs
        old_uris = [image.uri for image in self.images]

        # the uri to uri mapping
        uri2uri = {}

        if try_relative:
            assert not rebase_kwargs, "no other kwargs supported when try_relative=True"
            # test if project was moved (or mounted somewhere else...)
            prev_pth = GeneralTools.toPath(self.java_object.getPreviousURI()).getParent()
            proj_pth = GeneralTools.toPath(self.java_object.getURI()).getParent()
            if prev_pth and proj_pth and not prev_pth.equals(proj_pth):
                for old_uri in map(URI, old_uris):
                    img_pth = GeneralTools.toPath(old_uri)
                    new_pth = proj_pth.resolve(prev_pth.relativize(img_pth)).normalize()
                    if Files.exists(new_pth):
                        uri2uri[old_uri] = new_pth.normalize().toUri().normalize()

        else:
            # allow rebasing all image uris in a single call to rebase
            new_uris = self._image_provider.rebase(*old_uris, **rebase_kwargs)

            # build the java URI to URI mapping
            for old_uri, new_uri in zip(old_uris, new_uris):
                if new_uri is None:
                    continue
                elif ImageProvider.compare_uris(new_uri, old_uri):
                    continue
                uri2uri[URI(old_uri)] = URI(new_uri)

        # update uris if possible
        for image in self.images:
            image.java_object.updateServerURIs(uri2uri)

    @redirect(stderr=True, stdout=True)
    def remove_image(
        self,
        image_entry: Union[QuPathProjectImageEntry, int],
    ) -> None:
        """
        Delete an image from the QuPath project.

        Parameters
        ----------
        image_entry:
            the image entry to be removed
        """
        if isinstance(image_entry, int):
            image_entry = self.images[image_entry]
        if not isinstance(image_entry, QuPathProjectImageEntry):
            raise TypeError(
                f"expected QuPathProjectImageEntry, got: {type(image_entry).__name__!r}"
            )
        if self._readonly:
            raise OSError("project in readonly mode")
        try:
            self.java_object.removeImage(image_entry.java_object, False)
        finally:
            self._image_entries_proxy.refresh()
            self.save(images=False)

    @property
    def uri(self) -> str:
        """the uri identifying the project location"""
        return str(self.java_object.getURI().toString())

    # @property
    # def uri_previous(self) -> Optional[str]:
    #     """previous uri. potentially useful for re-associating"""
    #     uri = self.java_object.getPreviousURI()
    #     if uri is None:
    #         return None
    #     return str(uri.toString())

    @property
    def path_classes(self) -> Tuple[QuPathPathClass, ...]:
        """return path_classes stored in the project"""
        return tuple(map(QuPathPathClass.from_java, self.java_object.getPathClasses()))

    @path_classes.setter
    def path_classes(self, path_classes: Iterable[QuPathPathClass]):
        """to add path_classes reassign all path_classes here"""
        if self._readonly:
            raise AttributeError("project in readonly mode")
        pcs = [pc.java_object for pc in path_classes]
        self.java_object.setPathClasses(pcs)

    @property
    def path(self) -> pathlib.Path:
        """the path to the project root"""
        return pathlib.Path(str(self.java_object.getPath()))

    @redirect(stderr=True, stdout=True)
    def save(self, images=True) -> None:
        """flush changes in the project to disk

        (writes path_classes and project data)
        """
        if self._readonly:
            raise OSError("project in readonly mode")
        if images:
            for entry in self.images:
                entry.save()
        try:
            self.java_object.syncChanges()
        # convert java land exception
        except IOException:  # pragma: no cover
            raise OSError("occurred when trying to save the project")

    @property
    def name(self) -> str:
        """project name"""
        name = str(self.java_object.getName())
        if name.endswith("/project.qpproj"):
            name = name[:-15]
        return name

    def __repr__(self) -> str:
        # name = self.java_object.getNameFromURI(self.java_object.getURI())
        return f'{type(self).__name__}(path="{self._path}" mode="{self._mode}")'

    def _repr_html_(self) -> str:
        from paquo._repr import br
        from paquo._repr import div
        from paquo._repr import h3
        from paquo._repr import p
        from paquo._repr import repr_html
        from paquo._repr import span
        return div(
            h3(style={"margin-top": "0"}, text=f"Project: {self.name}"),
            p(
                span(text="path: ", style={"font-weight": "bold"}),
                span(text=str(self._path)),
                br(),
                span(text="mode: ", style={"font-weight": "bold"}),
                span(text=self._mode),
                style={"margin": "0.5em"},
            ),
            repr_html(self.images),
        )

    @property
    def timestamp_creation(self) -> int:
        """system time at creation in milliseconds"""
        return int(self.java_object.getCreationTimestamp())

    @property
    def timestamp_modification(self) -> int:
        """system time at modification in milliseconds"""
        return int(self.java_object.getModificationTimestamp())

    @property
    def version(self) -> Optional[str]:
        """the project version. should be identical to the qupath version"""
        # for older projects this returns null.
        # for newer projects this will be set to DefaultProject.LATEST_VERSION
        # which is just GeneralTools.getVersion()...
        version = self.java_object.getVersion()
        try:
            # note: this still implicitly requires a QuPath version that has GeneralTools
            #   which is probably fine...
            latest_version = GeneralTools.getVersion()
        except AttributeError:  # pragma: no cover
            latest_version = None
        # version is None, until we save a project to disk AND reload it!
        if version:
            return str(version)
        if latest_version:
            return str(latest_version)
        return None  # pragma: no cover

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self._readonly:
            self.save()



import io
import tempfile
from collections import namedtuple
from contextlib import redirect_stdout
from pathlib import Path

import pytest

from paquo.__main__ import main

_Output = namedtuple("_Output", "return_code stdout")


def run(func, argv1):
    f = io.StringIO()
    with redirect_stdout(f):
        return_code = func(argv1)
    return _Output(return_code, f.getvalue().rstrip())


@pytest.fixture(scope="module")
def project_with_data(svs_small):
    from shapely.geometry import Point
    from paquo.classes import QuPathPathClass
    from paquo.images import QuPathImageType
    from paquo.projects import QuPathProject

    with tempfile.TemporaryDirectory(prefix='paquo-') as tmp_path:
        with QuPathProject(tmp_path, mode='x') as qp:
            entry = qp.add_image(svs_small, image_type=QuPathImageType.BRIGHTFIELD_OTHER)
            entry.hierarchy.add_annotation(
                roi=Point(500, 500)
            )
            pcs = list(qp.path_classes)
            pcs.append(QuPathPathClass("myclass"))
            qp.path_classes = pcs
            entry.metadata["mykey"] = "myval"
            qp.save()
            yield qp.path


def test_no_args():
    assert main([]) == 1


def test_version():
    from paquo import __version__
    assert run(main, ['--version']) == (0, __version__)


def test_config_cmd(tmp_path):
    assert main(['config']) == 0  # shows help

    assert run(main, ['config', '-l']).stdout
    assert run(main, ['config', '-l', '--default']).stdout

    paquo_toml = Path(tmp_path) / ".paquo.toml"
    paquo_toml.touch()

    # error when folder does not exist
    with pytest.raises(SystemExit):
        assert run(main, ['config', '-l', '-o', str(tmp_path / "not-there")])
    # error when file exists
    assert run(main, ['config', '-l', '-o', str(tmp_path)]).return_code == 1
    # force allows overwrite
    assert run(main, ['config', '-l', '-o', str(tmp_path), '--force']).return_code == 0
    # allow showing all folders where you can store you paquo config
    assert run(main, ['config', '--search-tree']).return_code == 0


def test_list_cmd(project_with_data):
    assert run(main, ['list']).return_code == 0
    assert run(main, ['list', 'somewhere-non-existent']).return_code != 0
    assert run(main, ['list', str(project_with_data)]).return_code == 0


def test_create_cmd_errors(tmpdir):
    with tmpdir.as_cwd():
        # test help
        assert run(main, ['create']).return_code == 0
        # test classes class_colors mismatch
        assert run(main, ['create', 'test_project', '--classes', 'A', 'B', '--class-colors', '1']).return_code != 0
        # test duplicate classes
        assert run(main, ['create', 'test_project', '--classes', 'A', 'B', 'A']).return_code != 0
        # test image not there
        assert run(main, ['create', 'test_project', '--images', 'not-there.svs']).return_code != 0
        # test project is there
        p = Path(tmpdir) / "test_project"
        p.mkdir()
        qp = p / "project.qpproj"
        qp.touch()
        assert run(main, ['create', 'test_project', '--classes', 'A', 'B']).return_code != 0


def test_create_cmd_works(tmpdir, svs_small):
    with tmpdir.as_cwd():
        options = [
            'create',
            'test_project',
            '--classes', 'A',
            '--remove-default-classes',
            '--images', str(svs_small),
        ]
        assert run(main, options).return_code == 0
        assert (Path(tmpdir) / "test_project" / "project.qpproj").is_file()


def test_export_cmd(tmpdir, project_with_data):
    proj_path = project_with_data
    # help
    assert run(main, ['export']).return_code == 0
    # wrong index
    assert run(main, ['export', str(proj_path), '-i', '26']).return_code == 1
    # get index
    assert run(main, ['export', str(proj_path), '-i', '0']).return_code == 0
    # check pretty print
    assert run(main, ['export', str(proj_path), '-i', '0', '--pretty']).return_code == 0
    # check file output
    with tmpdir.as_cwd():
        assert run(main, ['export', str(proj_path), '-i', '0', '--pretty', '-o', 'data.geojson']).return_code == 0
        assert (Path(tmpdir) / "data.geojson").is_file()



from typing import Optional

from paquo.colors import ColorType
from paquo.colors import QuPathColor
from paquo.java import PathClass
from paquo.java import PathClassFactory
from paquo.java import String
from paquo.java import compatibility

__all__ = ['QuPathPathClass']


class QuPathPathClass:
    java_object: PathClass

    @classmethod
    def from_java(cls, path_class: PathClass) -> 'QuPathPathClass':
        """initialize a QuPathPathClass from its corresponding java PathClass"""
        if not isinstance(path_class, PathClass):
            raise TypeError("use PathClass(name='myclass') to instantiate")
        # keep type annotation requirements intact by providing
        # empty string, which is ignored when providing _java_path_class
        return cls('', _java_path_class=path_class)

    def __init__(self,
                 name: str,
                 color: Optional[ColorType] = None,
                 parent: Optional['QuPathPathClass'] = None,
                 **_kwargs) -> None:
        """create a QuPathPathClass

        The QuPathPathClasses are wrappers around singletons defined by their
        name and their ancestors. Internally there's only one java PathClass
        instance per ancestor chain + name. (The internal unique id is chaining
        names via ':', which is why ':' is an unsupported name character)

        Parameters
        ----------
        name:
            the name of your path class
        color:
            a color (r,g,b) or (r,g,b,a) with 0 < x < 255 or a QuPathColor
            if color is None, a color calculated from the name is used
        parent:
            the parent of the class

        Returns
        -------
        path_class:
            the QuPathPathClass
        """
        # internal: check if a java path class was already provided
        _java_path_class = _kwargs.pop('_java_path_class', None)
        if _java_path_class is not None:
            self.java_object = _java_path_class
            return

        # called by user
        if name is None:
            if parent is None:
                # note:
                #   parent=None for name=None creates the root node on the qupath java side
                #   will have to consider if we expose this here.
                raise NotImplementedError("creating Root PathClass is currently not supported")
            else:
                raise ValueError("cannot create derived QuPathPathClass with name=None")
        elif isinstance(name, str):
            if ":" in name or "\n" in name:
                raise ValueError("PathClass names cannot contain ':' or '\n'")
            name = String(name)
        else:
            raise TypeError(f"name requires type 'str' got '{type(name)}'")

        # get the parent class if requested
        java_parent = None
        if parent is not None:
            if not isinstance(parent, QuPathPathClass):
                raise TypeError("parent must be a QuPathPathClass")
            java_parent = parent.java_object

        # set the color if requested
        java_color = None
        if color is not None:
            java_color = QuPathColor.from_any(color).to_java_rgba()  # use rgba?

        if compatibility.supports_newer_addobject_and_pathclass():
            path_class = PathClass.getInstance(java_parent, name, java_color)
        else:
            path_class = PathClassFactory.getDerivedPathClass(java_parent, name, java_color)
        self.java_object = path_class

    @property
    def name(self) -> str:
        """the name of the class"""
        return str(self.java_object.getName())

    @property
    def id(self) -> str:
        """the unique identifier string of the class"""
        return str(self.java_object.toString())

    def __eq__(self, other) -> bool:
        # check if path classes are identical
        if not isinstance(other, QuPathPathClass):
            return False
        return int(self.java_object.compareTo(other.java_object)) == 0

    @property
    def parent(self) -> Optional['QuPathPathClass']:
        """the parent path class of this path class"""
        path_class = self.java_object.getParentClass()
        if path_class is None:
            return None
        return QuPathPathClass.from_java(path_class)

    @property
    def origin(self) -> 'QuPathPathClass':
        """the toplevel parent of this path class"""
        origin_class = self
        while origin_class.parent is not None:
            origin_class = origin_class.parent
        return origin_class

    def is_derived_from(self, parent_class: 'QuPathPathClass'):
        """is this class derived from the parent_class"""
        return self.java_object.isDerivedFrom(parent_class.java_object)

    def is_ancestor_of(self, child_class: 'QuPathPathClass'):
        """is this class an ancestor of the child_class"""
        return self.java_object.isAncestorOf(child_class.java_object)

    @property
    def color(self) -> Optional[QuPathColor]:
        """return the path color"""
        rgb = self.java_object.getColor()
        if rgb is None:
            return None
        return QuPathColor.from_java_rgb(rgb)

    @color.setter
    def color(self, rgb: Optional[ColorType]) -> None:
        """set the path color"""
        if rgb is not None:
            rgb = QuPathColor.from_any(rgb).to_java_rgb()  # maybe use argb?
        self.java_object.setColor(rgb)

    @property
    def is_valid(self) -> bool:
        """check if the path class is valid"""
        return bool(self.java_object.isValid())

    @property
    def is_derived_class(self) -> bool:
        """check if the class is a derived path class"""
        return bool(self.java_object.isDerivedClass())

    def __repr__(self) -> str:
        return f"QuPathPathClass('{self.id}')"



import json
import pathlib
import re
import warnings
import weakref
from collections.abc import MutableMapping
from copy import deepcopy
from enum import Enum
from pathlib import Path
from pathlib import PurePath
from pathlib import PurePosixPath
from pathlib import PureWindowsPath
from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Union
from urllib.parse import quote
from urllib.parse import urlsplit
from urllib.parse import urlunsplit

from paquo._logging import get_logger
from paquo._logging import redirect
from paquo._utils import cached_property
from paquo.hierarchy import QuPathPathObjectHierarchy
from paquo.java import URI
from paquo.java import BufferedImage
from paquo.java import DefaultProjectImageEntry
from paquo.java import File
from paquo.java import FileNotFoundException
from paquo.java import ImageType
from paquo.java import IOException
from paquo.java import NoSuchFileException
from paquo.java import PathIO
from paquo.java import String
from paquo.java import URISyntaxException
from paquo.java import compatibility

if TYPE_CHECKING:
    import paquo.projects

__all__ = [
    "ImageProvider",
    "QuPathImageType",
    "QuPathProjectImageEntry",
    "SimpleFileImageId",
]

_log = get_logger(__name__)


def __getattr__(name):
    if name == "SimpleURIImageProvider":
        warnings.warn(
            "SimpleURIImageProvider is deprecated. Please use ImageProvider",
            DeprecationWarning,
            stacklevel=2,
        )
        return ImageProvider
    raise AttributeError(name)


# [URI:java-python]
# NOTE: pathlib handles URIs a little different to QuPath's java URIs
#   having looked into it a little bit it seems neither are entirely
#   rfc3986 compliant because they both try to be permissive with
#   broken URIs...
#   For the sake of moving forward we go with the workarounds below.
#   This should all be replaced with rfc3986 compliant URI handling.
def _normalize_pathlib_uris(uri):
    """this will correctly unescape and normalize uri's received from pathlib.Path.as_uri()"""
    # https://docs.oracle.com/javase/7/docs/api/java/net/URI.html section Identities
    try:
        u = URI(uri)
    except URISyntaxException:
        try:
            s = urlsplit(uri)
            s = s._replace(path=quote(s.path))
            uri = urlunsplit(s)
        except ValueError:
            raise ValueError(f"uri not valid '{uri}'")
        else:
            u = URI(uri)
    scheme = u.getScheme()
    if scheme != "file":
        raise ValueError(f"uri unsupported scheme '{uri}'")
    host = u.getHost()
    path = str(u.getPath())
    if host:
        path = f"////{host}{path}"
    elif re.match("^//[^/]+/[a-zA-Z][$]/", path):
        path = f"//{path}"
    try:
        x = URI(
            scheme,
            u.getUserInfo(),
            None,
            u.getPort(),
            path,
            u.getQuery(),
            u.getFragment()
        )
    except URISyntaxException:
        raise ValueError(f"uri syntax error '{uri}'")
    return x


SimpleFileImageId = Union[str, pathlib.Path]


class ImageProvider:
    """Maps image ids to paths and paths to image ids."""

    class FilenamePathId(str):
        """an id that uses the filename as it's identifier"""
        def __eq__(self, other):
            return Path(self).name == Path(other).name

        def __hash__(self):
            return hash(Path(self).name)

        def __repr__(self):  # pragma: no cover
            p = Path(self)
            return f'FilenamePathId("{p.name}", parent="{p.parent}")'

    class URIString(str):
        """string uri's can differ in their string representation and still be identical"""
        # we need some way to normalize uris
        def __eq__(self, other):  # pragma: no cover
            return ImageProvider.compare_uris(self, other)
        __hash__ = str.__hash__  # fixme: this is not correct!

    def uri(self, image_id: SimpleFileImageId) -> Optional['URIString']:
        """accepts a path and returns a URIString"""
        if not isinstance(image_id, (Path, str, ImageProvider.FilenamePathId)):
            raise TypeError("image_id not of correct format")  # pragma: no cover
        if isinstance(image_id, str) and image_id.startswith("file:/"):
            # image_id is uri
            image_id = _normalize_pathlib_uris(image_id)
            return ImageProvider.URIString(image_id)
        img_path = pathlib.Path(image_id).absolute().resolve()
        if not img_path.is_file():
            return None
        return ImageProvider.URIString(img_path.as_uri())

    def id(self, uri: URIString) -> str:
        """accepts a uri string and returns a FilenamePathId"""
        if not isinstance(uri, (str, ImageProvider.URIString)):
            raise TypeError("uri not of correct format")  # pragma: no cover
        return ImageProvider.FilenamePathId(ImageProvider.path_from_uri(uri))

    def rebase(self, *uris: str, **kwargs) -> List[Optional[str]]:
        uri2uri = kwargs.pop('uri2uri', {})
        return [uri2uri.get(uri, None) for uri in uris]

    @staticmethod
    def path_from_uri(uri: str) -> PurePath:
        """
        Parses an URI representing a file system path into a Path.
        """
        # TODO: needs way more tests... See note [URI:java-python]
        java_uri = _normalize_pathlib_uris(uri)
        # test current scheme support
        if str(java_uri.getScheme()) != "file":
            raise NotImplementedError("paquo only supports file:/ URIs as of now")
        else:
            path_str = str(java_uri.getPath())

        host = java_uri.getHost()
        if host:
            path_str = f"//{host}{path_str}"

        # fixme: this should be replaced with something more reliable...
        # check if we encode a windows path
        if re.match(r"/[A-Z]:/[^/]", path_str):
            return PureWindowsPath(path_str[1:])
        elif re.match(r"//(?P<share>[^/]+)/(?P<directory>[^/]+)/", path_str):
            return PureWindowsPath(path_str)
        else:
            return PurePosixPath(path_str)

    @staticmethod
    def uri_from_path(path: PurePath) -> str:
        """
        Convert a python path object to an URI
        """
        # TODO: needs way more tests... See note [URI:java-python]
        if not path.is_absolute():
            raise ValueError("uri_from_path requires an absolute path")
        java_uri = str(_normalize_pathlib_uris(path.as_uri()).toString())
        # fixme: this should be replaced with a rfc3896 compliant solution...
        if re.match("file://([^/]|$)", java_uri):
            uri = f"file:////{java_uri[7:]}"  # network shares have redundant authority on the java side
        # vvv this would only be required if we wouldn't normalize the uri like above
        # elif re.match("file:///([^/]|$)", java_uri):
        #     uri = f"file:/{java_uri[8:]}"  # the local windows absolute paths don't
        else:
            uri = java_uri
        return uri

    @staticmethod
    def compare_uris(a: str, b: str) -> bool:
        """
        Test if two URIs point two the same resource
        """
        # TODO: needs way more tests... See note [URI:java-python]
        uri_a = _normalize_pathlib_uris(a)
        uri_b = _normalize_pathlib_uris(b)
        return bool(uri_a.equals(uri_b))


# noinspection PyPep8Naming
class _RecoveredReadOnlyImageServer:
    """internal. used to allow access to image server metadata recovered from project.qpproj"""

    # noinspection PyMethodParameters,PyPep8Naming
    class _FakeResolutionLevel:
        def __init__(self_, _lvl):
            self_._lvl = _lvl

        def getDownsample(self_):
            return self_._lvl['downsample']

        def getHeight(self_):
            return self_._lvl['height']

        def getWidth(self_):
            return self_._lvl['width']

    # noinspection PyMethodParameters,PyPep8Naming
    class _FakeMetadata:
        def __init__(self_, _metadata):
            self_._md = _metadata

        def nLevels(self_):
            return len(self_._md.get('levels', []))

        def getLevel(self_, idx):
            _rl = self_._md.get('levels')[idx]
            # noinspection PyProtectedMember
            return _RecoveredReadOnlyImageServer._FakeResolutionLevel(_rl)

    def __init__(self, entry_path: Path):
        server_json_f = Path(entry_path) / "server.json"
        with server_json_f.open('r') as f:
            self._metadata = json.load(f).get('metadata', {})

    def getWidth(self):
        return self._metadata['width']

    def getHeight(self):
        return self._metadata['height']

    def nChannels(self):
        return len(self._metadata['channels'])

    def nZSlices(self):
        return self._metadata['sizeZ']

    def nTimepoints(self):
        return self._metadata['sizeT']

    def getMetadata(self) -> Any:
        # fake the java metadata interface
        _md = deepcopy(self._metadata)
        # noinspection PyProtectedMember
        return _RecoveredReadOnlyImageServer._FakeMetadata(_md)


class _ProjectImageEntryMetadata(MutableMapping):
    """provides a python dict interface for image entry metadata"""

    def __init__(self, image: 'QuPathProjectImageEntry') -> None:
        self._image = image
        self._entry = image.java_object

    def __setitem__(self, k: str, v: str) -> None:
        # noinspection PyProtectedMember
        if self._image._readonly:
            raise AttributeError("project in readonly mode")
        if not isinstance(k, str):
            raise TypeError(f"key must be of type `str` got `{type(k)}`")
        if not isinstance(v, str):
            raise TypeError(f"value must be of type `str` got `{type(v)}`")
        self._entry.putMetadataValue(String(str(k)), String(str(v)))

    def __delitem__(self, k: str) -> None:
        # noinspection PyProtectedMember
        if self._image._readonly:
            raise AttributeError("project in readonly mode")
        if not isinstance(k, str):
            raise TypeError(f"key must be of type `str` got `{type(k)}`")
        self._entry.removeMetadataValue(String(str(k)))

    def __getitem__(self, k: str) -> str:
        if not isinstance(k, str):
            raise TypeError(f"key must be of type `str` got `{type(k)}`")
        v = self._entry.getMetadataValue(String(str(k)))
        if v is None:
            raise KeyError(f"'{k}' not in metadata")
        return str(v)

    def __len__(self) -> int:
        return int(self._entry.getMetadataKeys().size())

    def __iter__(self) -> Iterator[str]:
        return iter(map(str, self._entry.getMetadataKeys()))

    def __contains__(self, item):
        return bool(self._entry.containsMetadata(String(str(item))))

    def clear(self) -> None:
        # noinspection PyProtectedMember
        if self._image._readonly:
            raise AttributeError("project in readonly mode")
        self._entry.clearMetadata()

    def __repr__(self):
        return f"Metadata({repr(dict(self))})"


class _ImageDataProperties(MutableMapping):
    """provides a python dict interface for image data properties"""

    def __init__(self, image: 'QuPathProjectImageEntry') -> None:
        self._image = image
        # noinspection PyProtectedMember
        self._image_data = image._image_data

    def __setitem__(self, k: str, v: Any) -> None:
        # noinspection PyProtectedMember
        if self._image._readonly:
            raise AttributeError("project in readonly mode")
        if not isinstance(k, str):
            raise TypeError(f"key must be of type `str` got `{type(k)}`")
        self._image_data.setProperty(String(k), v)

    def __delitem__(self, k: str) -> None:
        # noinspection PyProtectedMember
        if self._image._readonly:
            raise AttributeError("project in readonly mode")
        if not isinstance(k, str):
            raise TypeError(f"key must be of type `str` got `{type(k)}`")
        self._image_data.removeProperty(String(k))

    def __getitem__(self, k: str) -> Any:
        if not isinstance(k, str):
            raise TypeError(f"key must be of type `str` got `{type(k)}`")
        if k not in self:
            raise KeyError(f"'{k}' not in metadata")
        v = self._image_data.getProperty(String(k))
        return v

    def __contains__(self, item: Any) -> bool:
        if not isinstance(item, str):
            return False
        return bool(
            self._image_data.getProperties().containsKey(String(item))
        )

    def __len__(self) -> int:
        return int(self._image_data.getProperties().size())

    def __iter__(self) -> Iterator[str]:
        return iter(map(str, dict(self._image_data.getProperties())))

    def __repr__(self):
        return f"Properties({repr(dict(self))})"


# note: this could just be autogenerated by inspecting the ImageType
#   but it's better to be explicit so that all values are defined here
class QuPathImageType(str, Enum):
    """Enum representing image types"""
    java_enum: ImageType

    def __new__(cls, value: str, java_enum: ImageType):
        # noinspection PyArgumentList
        obj = super().__new__(cls, value)
        obj._value_ = value
        obj.java_enum = java_enum
        return obj

    @classmethod
    def from_java(cls, java_enum) -> 'QuPathImageType':
        """internal for converting from java to python"""
        for value in cls.__members__.values():
            if value.java_enum == java_enum:
                return value
        raise ValueError("unsupported java_enum")  # pragma: no cover

    # Brightfield image with hematoxylin and DAB stains.
    BRIGHTFIELD_H_DAB = ("Brightfield (H-DAB)", ImageType.BRIGHTFIELD_H_DAB)
    # Brightfield image with hematoxylin and eosin stains.
    BRIGHTFIELD_H_E = ("Brightfield (H&E)", ImageType.BRIGHTFIELD_H_E)
    # Brightfield image with any stains.
    BRIGHTFIELD_OTHER = ("Brightfield (other)", ImageType.BRIGHTFIELD_OTHER)
    # Fluorescence image.
    FLUORESCENCE = ("Fluorescence", ImageType.FLUORESCENCE)
    # Other image type, not covered by any of the alternatives above.
    OTHER = ("Other", ImageType.OTHER)
    # Image type has not been set.
    UNSET = ("Not set", ImageType.UNSET)


class QuPathProjectImageEntry:
    java_object: DefaultProjectImageEntry

    def __init__(self, entry: DefaultProjectImageEntry,
                 *, _project_ref: Optional['paquo.projects.QuPathProject'] = None) -> None:
        """Wrapper for qupath image entries

        this is normally not instantiated by the user
        """
        if not isinstance(entry, DefaultProjectImageEntry):
            raise ValueError("don't instantiate directly. use `QuPathProject.add_image`")
        self.java_object = entry
        self._project_ref = weakref.ref(_project_ref) if _project_ref else lambda: None
        self._metadata = _ProjectImageEntryMetadata(self)

    @property
    def _readonly(self):
        p = self._project_ref()
        return getattr(p, "_readonly", False) if p else True

    @cached_property
    def _image_data(self):
        with redirect(stdout=True, stderr=True):
            try:
                return self.java_object.readImageData()
            # from java land
            except IOException:  # pragma: no cover
                image_data_fn = self.entry_path / "data.qpdata"
                try:
                    image_data = PathIO.readImageData(
                        File(str(image_data_fn)),
                        None, None, BufferedImage
                    )
                except (FileNotFoundException, NoSuchFileException):
                    raise FileNotFoundError("image_data missing")
                return image_data

    @cached_property
    def _properties(self):
        return _ImageDataProperties(self)

    @cached_property
    def _image_server(self):
        server = self._image_data.getServer()
        if not server:
            _log.warning("recovering readonly from server.json")
            try:
                server = _RecoveredReadOnlyImageServer(self.entry_path)
            except FileNotFoundError:
                if not compatibility.supports_image_server_recovery():
                    raise RuntimeError("QuPath < 0.2.0 is not guaranteed to write server.json")
                raise
        return server

    @property
    def entry_id(self) -> str:
        """the unique image entry id"""
        return str(self.java_object.getID())

    @property
    def entry_path(self) -> Path:
        """path to the image directory"""
        return Path(str(self.java_object.getEntryPath().toString()))

    @property
    def image_name(self) -> str:
        """the image entry name"""
        return str(self.java_object.getImageName())

    @image_name.setter
    def image_name(self, name: str) -> None:
        if self._readonly:
            raise AttributeError("project in readonly mode")
        self.java_object.setImageName(String(name))

    # remove until there's a good use case for this...
    # @property
    # def image_name_original(self) -> Optional[str]:
    #     """original name in case the user has changed the image name"""
    #     org_name = self.java_object.getOriginalImageName()
    #     return str(org_name) if org_name else None

    @property
    def image_type(self) -> QuPathImageType:
        """image type"""
        return QuPathImageType.from_java(self._image_data.getImageType())

    @image_type.setter
    def image_type(self, value: QuPathImageType) -> None:
        if self._readonly:
            raise AttributeError("project in readonly mode")
        if not isinstance(value, QuPathImageType):
            raise TypeError("requires a QuPathImageType enum")
        self._image_data.setImageType(value.java_enum)

    @property
    def description(self) -> str:
        """free text describing the image"""
        text = self.java_object.getDescription()
        if text is None:
            return ""
        return str(text)

    @description.setter
    def description(self, text: str) -> None:
        if self._readonly:
            raise AttributeError("project in readonly mode")
        self.java_object.setDescription(text)

    @property
    def width(self):
        """image width in pixels"""
        return int(self._image_server.getWidth())

    @property
    def height(self):
        """image height in pixels"""
        return int(self._image_server.getHeight())

    @property
    def num_channels(self):
        """number of channels in the image"""
        return int(self._image_server.nChannels())

    @property
    def num_z_slices(self):
        """number of z_slices in the image"""
        return int(self._image_server.nZSlices())

    @property
    def num_timepoints(self):
        """number of time points in the image"""
        return int(self._image_server.nTimepoints())

    @cached_property
    def downsample_levels(self) -> List[Dict[str, float]]:
        """downsample levels provided by the image

        Notes
        -----
        The available downsample levels can differ dependent
        on which image backend is used by QuPath
        """
        with redirect(stdout=True, stderr=True):
            md = self._image_server.getMetadata()
        levels = []
        for level in range(int(md.nLevels())):
            resolution_level = md.getLevel(level)
            levels.append({
                'downsample': float(resolution_level.getDownsample()),
                'width': int(resolution_level.getWidth()),
                'height': int(resolution_level.getHeight()),
            })
        return levels

    @property
    def metadata(self) -> _ProjectImageEntryMetadata:
        """the metadata stored on the image as dict-like proxy"""
        return self._metadata

    @metadata.setter
    def metadata(self, value: dict) -> None:
        if self._readonly:
            raise AttributeError("project in readonly mode")
        self._metadata.clear()
        self._metadata.update(value)

    @property
    def properties(self):
        """the properties stored in the image data as a dict-like proxy"""
        return self._properties

    @properties.setter
    def properties(self, value):
        if self._readonly:
            raise AttributeError("project in readonly mode")
        self._properties.clear()
        self._properties.update(value)

    @cached_property
    def hierarchy(self) -> QuPathPathObjectHierarchy:
        """the image entry hierarchy. it contains all annotations"""
        if self._readonly:
            h = self.java_object.readHierarchy()
        else:
            try:
                h = self._image_data.getHierarchy()
            except OSError:
                _log.warning("could not open image data. loading annotation hierarchy from project.")
                h = self.java_object.readHierarchy()

        return QuPathPathObjectHierarchy(h, readonly=self._readonly, image_name=self.image_name)

    def __repr__(self):
        return f"ImageEntry(image_name='{self.image_name}')"

    def _repr_html_(self, compact=False, index=0):
        from base64 import b64encode

        from paquo._repr import br
        from paquo._repr import div
        from paquo._repr import h4
        from paquo._repr import img
        from paquo._repr import p
        from paquo._repr import span

        img_css = {
            "max-width": "100px",
            "max-height": "100px",
            "border": "1px solid",
            "margin": "auto",
        }
        header_css = {
            "position": "absolute",
            "top": "-1.6em",
            "width": "100px",
            "overflow": "hidden",
            "text-overflow": "ellipsis",
            "font-size": "0.75em",
        }
        container_css = {
            "display": "flex",
            "align-items": "center",
            "justify-content": "center",
            "position": "relative",
            "width": "100px",
            "height": "100px",
            "background": "#ddd",
            "margin": "2px",
        }

        try:
            with (self.entry_path / "thumbnail.jpg").open(mode="rb") as f:
                data = b64encode(f.read()).decode('utf-8')
        except FileNotFoundError:  # pragma: no cover
            image = span(style={"font-size": "3em"}, text="?")
        else:
            image = img(title=self.image_name,
                        src=f"data:image/jpeg;base64,{data}",
                        style=img_css)
        if compact:
            container_css["margin-top"] = "1em"
            return div(
                span(text=f"[{index}]\xa0{self.image_name}", style=header_css),
                image,
                style=container_css
            )

        try:
            uri = self.uri[5:]
        except RuntimeError as err:  # pragma: no cover
            uri = f"N/A ({err})"
        return div(
            h4(text=f"Image: {self.image_name}", style={"margin-top": "0"}),
            p(
                span(text="path: ", style={"font-weight": "bold"}),
                span(text=uri),
                br(),
                span(text="type: ", style={"font-weight": "bold"}),
                span(text=str(self.image_type.value)),
                style={"margin": "0.5em"},
            ),
            div(
                image,
                style=container_css,
            )
        )

    @property
    def uri(self):
        """the image entry uri"""
        uris = self.java_object.getServerURIs()
        if len(uris) == 0:
            raise RuntimeError("no server")  # pragma: no cover
        elif len(uris) > 1:
            raise NotImplementedError("unsupported in paquo as of now")
        return str(uris[0].toString())

    def is_readable(self) -> bool:
        """check if the image file is readable"""
        concrete_path = Path(ImageProvider.path_from_uri(self.uri))
        return concrete_path.is_file()

    def is_changed(self) -> bool:
        """check if image_data is changed

        Raises
        ------
        IOError
            if image_data can't be read

        """
        return bool(self._image_data.isChanged())

    def save(self):
        """save image entry"""
        with redirect(stdout=True, stderr=True):
            if self._readonly:
                raise OSError("project in readonly mode")
            if self.is_readable():
                if self.is_changed():
                    self.java_object.saveImageData(self._image_data)
                    _log.info(f"Changes saved for '{self.image_name}'")
                else:
                    _log.info(f"Saving skipped for '{self.image_name}': no changes")
            else:
                _log.warning(f"Saving skipped for '{self.image_name}': uri '{self.uri}' not reachable")
